//
// Created by R on 9/25/23.
//

#include "PIMD.h"
#include "Vibro.h"

PIMD::PIMD(PISampler &pisampler, ProtocolComm &pc, UI &ui, IndicationStateMachine &ism,
           ThresholdDetectorSM &td, HAL& hal) :
        _op_state(STATE_INIT),
        _pi(pisampler),
        _pc(pc),
        _ui(ui),
        _params(hal.params),
        _auto_offset_sm(_pi.get_sampling_configuration()),
        _td(td),
        _ref_sample(_pi.get_sampling_configuration()),
        _ism(ism),
        _numOperationalTransitions(0),
        _nextRefSampleConsumptionUs(0),
        _bm(hal.bm),
        _statusStruct(SystemStatus_init_default),
        _samplerateCounter({0, 0}),
        _indicate_battery_on_startup(true),
        _vibro(hal.vibro),
        _tm(hal.tm),
        _offset_adjustment_end_timestamp_us(0),
        _audio(hal.audio),
        _top_clipping_counter(0)
{
    memset(&_battery_level_events, 0, sizeof(_battery_level_events));
}

void PIMD::step() {
    if (_pi.samplingConfigurationChanged()) {
        _ref_sample.setSamplingConfiguration(_pi.get_sampling_configuration());
        _auto_offset_sm.setSamplingConfiguration(_pi.get_sampling_configuration());
    }
    switch (_op_state) {
        case STATE_INIT:
            state_step_init();
            break;
        case STATE_INITIALIZATION_AUDIO_PLAYBACK:
            if (_ui.getAudio().isPlaybackFinished()) {
                auto next_state = STATE_START_REF_SAMPLE_ACQUISITION;
                //don't need to enter auto offset adjustment if offset is manually controlled
                if (!_pi.get_sampling_configuration().frontend_manual_offset_control) {
                    next_state = STATE_START_AUTO_OFFSET_ADJUSTMENT;
                }
                state_transition(next_state);
            }
            break;
        case STATE_INDICATE_BATTERY_LEVEL:
            state_step_indicate_battery_level();
            break;
        case STATE_START_AUTO_OFFSET_ADJUSTMENT:
            _auto_offset_sm.setSamplingConfiguration(_pi.get_sampling_configuration());
            _auto_offset_sm.init();
            state_transition(STATE_AUTO_OFFSET_ADJUSTMENT);
            break;
        case STATE_AUTO_OFFSET_ADJUSTMENT:
            state_step_offset_adjustment();
            break;
        case STATE_START_REF_SAMPLE_ACQUISITION:
            _pi.clearDecaySampleHistory();
            _ref_sample.reset();
            _td.resetState();
            state_transition(STATE_REF_SAMPLE_ACQUISITION);
            break;
        case STATE_REF_SAMPLE_ACQUISITION:
            state_step_ref_sample_acquisition();
            break;
        case STATE_THRESHOLD_INITIALIZATION:
            state_step_threshold_initialization();
            break;

        case STATE_OPERATIONAL:
            state_step_operational();
            break;
        case STATE_ERROR_AUTO_OFFSET_FAILED:
        case STATE_ERROR_REF_SAMPLE_ACQUISITION_FAILED:
            assert(false);
            break;
        default:
            assert(false);
            break;
    }

    _pc.communicate();
    process_protocol_commands();
    periodic_send_reference_measurement();
    periodic_battery_monitor_measurement();
    periodic_temperature_measurement();
    periodic_send_system_status();
    _ui.process_events();
    process_user_commands();
}

void PIMD::state_step_init() {
    _nextSystemStatusSendTimeUs = time_us_64() + SYSTEM_STATUS_TRANSMIT_PERIOD_US;
    if (_params.readParameters()) {
        _ui.set_sensitivity_setting(_params.parameters.ui_sensitivity);
        _ui.set_volume_setting(_params.parameters.ui_volume);
        _ui.set_vibro_power_setting(_params.parameters.ui_vibro_power);
        _ui.set_backlight_setting(_params.parameters.ui_backlight_enabled);
        _ui.set_brightness_setting(_params.parameters.ui_brightness);
        _ui.set_detection_mode_setting(_params.parameters.ui_detection_mode);
        if (_params.parameters.has_sampling_configuration) {
            _pi.set_sampling_configuration(_params.parameters.md_sampling_configuration);
            _ref_sample.setSamplingConfiguration(_params.parameters.md_sampling_configuration);
        }

        if (_params.parameters.has_threshold_configuration) {
            _td.setConfiguration(_params.parameters.md_threshold_configuration);
        }

        applyUISettings();
    }
    if (_indicate_battery_on_startup) {

        _state_after_battery_indication = STATE_INITIALIZATION_AUDIO_PLAYBACK;
        state_transition(STATE_INDICATE_BATTERY_LEVEL);
    } else {
        state_transition(STATE_INITIALIZATION_AUDIO_PLAYBACK);
    }
    _td.updateSamplingFrequency(_pi.get_sample_frequency_hz());
}

void PIMD::state_step_offset_adjustment() {
    decay_sample_t sample;
    if (_pi.getDecaySampleQueue().try_remove(sample)) {
        float dc;
        auto res = _auto_offset_sm.step(sample, dc);
        _pi.set_offset_duty_cycle(dc);
        if (res == AutoOffsetAdjustSM::FINISHED) {
            state_transition(STATE_START_REF_SAMPLE_ACQUISITION);
            _offset_adjustment_end_timestamp_us = time_us_64();
        }

        _pc.send_detection_measurement(sample);
    }
}

void PIMD::state_step_ref_sample_acquisition() {

    decay_sample_t sample;
    if (_pi.getDecaySampleQueue().try_remove(sample)) {
        _ref_sample.consumePotentialReferenceSample(sample);
        if (_ref_sample.isValid()) {
            _td.resetState();
            state_transition(STATE_THRESHOLD_INITIALIZATION);
        }
    }
}

void PIMD::state_step_operational() {
    decay_sample_t sample;
    if (_pi.getDecaySampleQueue().try_remove(sample)) {
        if (_nextRefSampleConsumptionUs < time_us_64()) {
            _ref_sample.consumePotentialReferenceSample(sample);
            _nextRefSampleConsumptionUs = time_us_64() + _pi.get_sampling_configuration().ref_sample_update_period_us;
        }
        auto stats = SampleAnalyzer::sample_compute_stats(
                sample, _ref_sample.getReferenceDecaySample(),
                _pi.get_sampling_configuration().num_samples_analyze,
                _td.getConfiguration().compute_expensive_stats);

        float amplitude;
        if (_td.consumeSample(sample, stats, _ref_sample.getReferenceDecaySample(), _ref_sample.getReferenceDecaySampleStats(), amplitude)) {
            if (!_ui.in_menu())
                _ism.indicateDetection(amplitude, 0.0f);
        }

        _samplerateCounter.numSamplesProcessed++;
        _pc.send_detection_measurement(sample, &stats);

        if (stats.top_clipping_detected && !_pi.get_sampling_configuration().frontend_manual_offset_control) {
            _top_clipping_counter++;
        } else {
            _top_clipping_counter = 0;
        }
        if (_top_clipping_counter >= _pi.get_sample_frequency_hz() * 5) {
            _audio.playSystemSound(PWMAudio::SOUND_SIGNAL_CLIPPING_DETECTED);
            _audio.waitSystemSoundPlaybackFinished();
            state_transition(STATE_INITIALIZATION_AUDIO_PLAYBACK);
        }
    }
}

void PIMD::state_transition(PIMD::op_state_t new_state) {
    _pc.log_message("PIMD state transition: %d -> %d", _op_state, new_state);
    _op_state = new_state;
    switch (new_state) {
        case STATE_INITIALIZATION_AUDIO_PLAYBACK:
            _ui.getAudio().playSystemSound(PWMAudio::SOUND_INITIALIZING);
            _ui.set_led_bar_level(0);
            sleep_ms(10); // to allow audio core to start playback
            break;
        case STATE_START_AUTO_OFFSET_ADJUSTMENT:
            _ui.getAudio().playMelody(MELODY_INIT);
            break;
        case STATE_OPERATIONAL:
            _numOperationalTransitions++;
            _top_clipping_counter = 0;
            _audio.playMelody(MELODY_CONFIRM);
            _audio.waitMelodyPlaybackFinished();
            _audio.setPlaybackState(PWMAudio::PLAYBACK_DETECTION);
            _samplerateCounter.numSamplesProcessed = 0;
            _samplerateCounter.samplerateStartTimeUs = time_us_64();
            break;
        case STATE_INDICATE_BATTERY_LEVEL: {
            auto bs = _bm.read_battery_charge();
            auto lev = std::clamp(bs.battery_soc / 10, 1, 9);
            _ui.set_led_bar_level(lev, UI::COLOR_MAP_RED_TO_GREEN_SAME);
            _audio.setPlaybackState(PWMAudio::PLAYBACK_STOPPED);

            auto snd = battery_soc_to_level(bs.battery_soc);
            _ui.getAudio().playSystemSound(snd);
            _battery_level_events.prev_indicated_level = snd;
            sleep_ms(10); // to allow audio core to start playback
        }
            break;
        default:
            break;
    }
}

void PIMD::process_protocol_commands() {
    ProtocolComm::protocol_command_t cmd;
    if (_pc.getProtocolCommandQueue().try_remove(cmd)) {
        switch (cmd) {
            case ProtocolComm::PC_REQUEST_REFERENCE_SAMPLE_ACQUISITION:
                state_transition(STATE_START_REF_SAMPLE_ACQUISITION);
                break;
            case ProtocolComm::PC_REQUEST_AUTOMATIC_OFFSET_ADJUSTMENT:
                state_transition(STATE_START_AUTO_OFFSET_ADJUSTMENT);
                break;
            case ProtocolComm::PC_REQUEST_NOISE_FLOOR_MEASUREMENT:
                _td.resetState();
                state_transition(STATE_THRESHOLD_INITIALIZATION);
                break;
        }
    }

}

void PIMD::periodic_send_reference_measurement() {
    static uint64_t next_send_time = 0;
    if (time_us_64() > next_send_time && _ref_sample.isValid()) {
        next_send_time = time_us_64() + 1000000;
        auto stats = _ref_sample.getReferenceDecaySampleStats();
        _pc.send_reference_measurement(_ref_sample.getReferenceDecaySample(), &stats);
        _pc.send_noise_floor_status_report();
    }
}

void PIMD::periodic_battery_monitor_measurement() {
    //read the battery more frequently, update the UI battery level less frequently
    static uint64_t next_measure_time = 0;
    auto ts = time_us_64();
    if (ts > next_measure_time) {
        next_measure_time = ts + 100000;
        _bm.update_battery_charge();
    }

    static uint64_t next_ui_update_time = 0;
    if (ts > next_ui_update_time) {
        next_ui_update_time = ts + 1000000;
        _ui.set_battery_soc(_bm.read_battery_charge().battery_soc);
    }

    static uint64_t next_mon_time = time_us_64() + 1000000;
    if (ts > next_mon_time) {
        next_mon_time = ts + 1000000;
        auto bs = _bm.read_battery_charge();
        auto level = battery_soc_to_level(bs.battery_soc);
        if (level == PWMAudio::SOUND_BATTERY_LEVEL_LOW) {
            // indicate low battery level when first reached, unless it has already been indicated
            auto prev_level = _battery_level_events.prev_indicated_level;
            if (prev_level != PWMAudio::SOUND_BATTERY_LEVEL_LOW &&
                prev_level != PWMAudio::SOUND_BATTERY_LEVEL_CRITICAL && _battery_level_events.num_indicated_low == 0) {
                _battery_level_events.num_indicated_low++;
                state_transition(STATE_INDICATE_BATTERY_LEVEL);
            }
        } else if (level == PWMAudio::SOUND_BATTERY_LEVEL_CRITICAL) {
            if (time_us_64() > _battery_level_events.next_crit_indication_ts) {
                _battery_level_events.num_indicated_critical++;
                _battery_level_events.next_crit_indication_ts = time_us_64() + 60 * 1000 * 1000;
                state_transition(STATE_INDICATE_BATTERY_LEVEL);
            }
        }
    }
}

void PIMD::process_user_commands() {
    UI::ui_cmd_t ui_cmd;
    if (_ui.get_pending_command(ui_cmd)) {
        switch (ui_cmd) {
            case UI::CMD_PERFORM_ZERO:
                state_transition(STATE_START_REF_SAMPLE_ACQUISITION);
                _audio.playSystemSound(PWMAudio::SOUND_REF_SAMPLE_UPDATED);
                break;
            case UI::CMD_UPDATE_SETTINGS:
                _params.parameters.ui_volume = _ui.get_volume_setting();
                _params.parameters.ui_vibro_power = _ui.get_vibro_setting();
                _params.parameters.ui_sensitivity = _ui.get_sensitivity_setting();
                _params.parameters.ui_brightness = _ui.get_brightness_setting();
                _params.parameters.ui_backlight_enabled = _ui.get_backlight_setting();
                _params.parameters.ui_detection_mode = _ui.get_detection_mode_setting();
                _params.writeParameters();
                sleep_ms(10);
                applyUISettings();
                _pc.send_threshold_configuration();
                _audio.playMelody(MELODY_CONFIRM);
                _audio.waitMelodyPlaybackFinished();
                break;
            case UI::CMD_INDICATE_VOLUME_CHANGE:
                _ui.updateVolume();
                _audio.playMelody(MELODY_SHORT_BEEP);
                break;
            case UI::CMD_INDICATE_VIBRO_POWER_CHANGE:
                _ui.updateVibroPower();
                _vibro.vibrate(1.0f);
                sleep_ms(100);
                _vibro.vibrate(0.0f);
                break;
            case UI::CMD_INDICATE_POWER_OFF:
                _audio.setPlaybackState(PWMAudio::PLAYBACK_STOPPED);
                _audio.playMelody(MELODY_POWER_OFF);
                _vibro.set_max_power(0.0f);
                _ui.set_led_bar_level(0);
                break;
            case UI::CMD_INDICATE_BATTERY_LEVEL:
                //only allow checking battery level when operational
                if (_op_state == STATE_OPERATIONAL)
                    state_transition(STATE_INDICATE_BATTERY_LEVEL);
                break;
            case UI::CMD_ENTER_MENU:
                _audio.setPlaybackState(PWMAudio::PLAYBACK_STOPPED);
                break;
            case UI::CMD_EXIT_MENU:
                _audio.setPlaybackState(PWMAudio::PLAYBACK_DETECTION);
                break;
        }
    }
}

void PIMD::state_step_threshold_initialization() {
    decay_sample_t sample;
    if (_pi.getDecaySampleQueue().try_remove(sample)) {
        _ref_sample.consumePotentialReferenceSample(sample);

        float tmp;
        auto stats = SampleAnalyzer::sample_compute_stats(sample, _ref_sample.getReferenceDecaySample(),
                                                          _pi.get_sampling_configuration().num_samples_analyze, false);

        //after offset adjustment, need to wait a few seconds until the noise floor settles
        //as a workaround- just wait a few seconds after the offset adjustment is finished
        bool enoughTimeAfterOffsetAdjustment = time_us_64() - _offset_adjustment_end_timestamp_us > 5 * 1000 * 1000;
        if (enoughTimeAfterOffsetAdjustment && _ref_sample.isValid() && _td.consumeSample(sample, stats, _ref_sample.getReferenceDecaySample(),
                               _ref_sample.getReferenceDecaySampleStats(), tmp)) {
            if (_indicate_battery_on_startup) {
                _indicate_battery_on_startup = false;
                state_transition(STATE_INDICATE_BATTERY_LEVEL);
            } else {
                state_transition(STATE_OPERATIONAL);
            }
            _pc.send_noise_floor_status_report();
        }
        _pc.send_detection_measurement(sample, &stats);
    }
}

void PIMD::applyUISettings() {
    _td.setSensitivityFromUserSetting(_ui.get_sensitivity_setting());
    _td.setDetectionModeFromUserSetting(_ui.get_detection_mode_setting());
    _ui.updateVolume();
    _ui.updateVibroPower();
}

void PIMD::update_system_status() {
    auto t = time_us_64();
    auto dt_us = t - _samplerateCounter.samplerateStartTimeUs;
    //update status member
    _statusStruct.processed_sample_rate = (uint64_t)_samplerateCounter.numSamplesProcessed * 1000000 / dt_us;
    _statusStruct.adc_read_sample_rate = (uint64_t)_pi.getNumADCDecayCurvesRead() * 1000000 / dt_us;
    _statusStruct.mcu_temperature_c = _tm.read_temperature_C().mcu_temperature_C;
    _statusStruct.timestamp_us = t;
    auto batstate = _bm.read_battery_charge();
    _statusStruct.battery_voltage = batstate.battery_voltage;
    _statusStruct.battery_soc = batstate.battery_soc;
    //reset the counter state
    _samplerateCounter.samplerateStartTimeUs = t;
    _samplerateCounter.numSamplesProcessed = 0;
    _pi.setNumADCDecayCurvesRead(0);
    _statusStruct.headphones_detected = _ui.getAudio().headphonesDetected();
}

void PIMD::periodic_send_system_status() {
    auto t = time_us_64();
    if (t > _nextSystemStatusSendTimeUs) {
        update_system_status();
        _pc.send_system_status(_statusStruct);
        _nextSystemStatusSendTimeUs = t + SYSTEM_STATUS_TRANSMIT_PERIOD_US;
    }
}

void PIMD::state_step_indicate_battery_level() {
    if (!_ui.getAudio().isSystemSoundPlaying()) {
        state_transition(_state_after_battery_indication);
        if (_indicate_battery_on_startup) {
            _indicate_battery_on_startup = false;
            _state_after_battery_indication = STATE_OPERATIONAL;
        }
        _audio.setPlaybackState(PWMAudio::PLAYBACK_DETECTION);
    }
}

PWMAudio::system_sound_key_t PIMD::battery_soc_to_level(uint8_t soc) {
    if (soc > 75) {
        return PWMAudio::SOUND_BATTERY_LEVEL_HIGH;
    } else if (soc > 33) {
        return PWMAudio::SOUND_BATTERY_LEVEL_MEDIUM;
    } else if (soc > 10) {
        return PWMAudio::SOUND_BATTERY_LEVEL_LOW;
    } else {
        return PWMAudio::SOUND_BATTERY_LEVEL_CRITICAL;
    }
}

void PIMD::periodic_temperature_measurement() {
    static uint64_t next_measure_time = 0;
    auto ts = time_us_64();
    if (ts > next_measure_time) {
        next_measure_time = ts + 50 * 1000;
        _tm.update_temperature();
    }
}
