//
// Created by R on 3/4/23.
//

#include "ProtocolComm.h"

#include <sys/types.h>
#include <cstdio>
#include <cstdarg>
#include <pico/stdio/driver.h>
#include <pico/stdio_usb.h>
#include <pb_common.h>
#include <pb_encode.h>
#include <pb_decode.h>
#include <algorithm>
#include "hardware/uart.h"
#include "pi.pb.h"
#include "nanoparser.h"
#include <pico/unique_id.h>
#include <git.h>
#include <hardware/watchdog.h>

static ProtocolComm* PC_INST=nullptr;
void on_uart_rx() {
    PC_INST->onUartRXIRQ();
}

ProtocolComm::ProtocolComm(PISampler &pi, ThresholdDetectorSM &td, HAL &hal)
        : _pi(pi), _td(td), _params(hal.params), _bm(hal.bm), _vibro(hal.vibro), _numDetectionMeasurements(0),
          _uartInitialized(false),
      _uartInst(nullptr) 
{
  parser_init_msg(&_parser_received_msg);
  PC_INST = this;
}

size_t ProtocolComm::write_bytes(const uint8_t *data, size_t len) {
    auto usb_written = fwrite(data, sizeof(uint8_t), len, stdout);
    if (_uartInitialized) {
        uart_write_blocking(_uartInst, data, len);
        return len;
    }
    return usb_written;
}

size_t ProtocolComm::read_bytes(uint8_t *data, size_t max_len) {
    auto ret = stdio_usb.in_chars((char*)data, (int)max_len);
    if (ret != PICO_ERROR_NO_DATA) {
        return ret;
    } else {
        uint nread=0;
        while (!_uartRXBuffer.is_empty()) {
            data[nread] = _uartRXBuffer.remove_blocking();
            nread++;
        }
        return nread;
    }
    return 0;
}

void ProtocolComm::send_log_message(const char *msg, size_t len) {
    LogEntry le = LogEntry_init_default;
    strncpy(le.message, msg, std::min(sizeof(le.message), len));
    size_t framed_size=0;
    if (PARSER_FRAME_MESSAGE(&le, LogEntry, _outmsgbuf, sizeof(_outmsgbuf), &framed_size) == PARSER_RESULT_MSG_OK) {
        write_bytes(_outmsgbuf, framed_size);
    }
}

void ProtocolComm::send_board_identification_message() {
    BoardIdentification msg = BoardIdentification_init_default;
    pico_get_unique_board_id_string(msg.board_id, sizeof(msg.board_id));

    strncpy(msg.commit_hash, git_CommitSHA1(), sizeof(msg.commit_hash));

    strncpy(msg.version, git_Describe(), sizeof(msg.version));
    strncpy(msg.board_type, PIFW_BOARD, sizeof(PIFW_BOARD));
    size_t framed_size=0;
    if (PARSER_FRAME_MESSAGE(&msg, BoardIdentification , _outmsgbuf,
                             sizeof(_outmsgbuf), &framed_size) == PARSER_RESULT_MSG_OK) {
        write_bytes(_outmsgbuf, framed_size);
    }

}

void ProtocolComm::communicate() {
    auto recv_len = read_bytes(_inmsgbuf, sizeof(_inmsgbuf));

    if (recv_len > 0) {
        parser_result_t res;
        for (int i=0; i < recv_len; i++) {
            res = parser_parse_byte(&_parser_received_msg, _inmsgbuf[i]);
            switch (res) {
                case PARSER_RESULT_MSG_OK:
                    process_received_message(&_parser_received_msg);
                    break;
                case PARSER_RESULT_INVALID_LENGTH:
                    log_message("received msg with invalid length");
                    break;
                case PARSER_RESULT_INVALID_CRC:
                    log_message("received message with invalid crc");
                    break;
                default:
                    break;
            }
            if (res != PARSER_RESULT_NONE)
                parser_init_msg(&_parser_received_msg);
        }
    }
}


void ProtocolComm::process_received_message(parser_msg_t *msg) {
    pb_istream_t is = pb_istream_from_buffer(msg->payload, msg->length);
    switch (msg->msgid) {
        case BoardIdentificationRequest_msgid: {
            send_board_identification_message();
        } break;
        case SamplingConfiguration_msgid: {
            SamplingConfiguration sc = SamplingConfiguration_init_default;
            if (pb_decode(&is, SamplingConfiguration_fields, &sc)) {
                handle_sampling_configuration(sc);
            }
        } break;
        case SamplingConfigurationRequest_msgid: {
            send_sampling_configuration();
        } break;
        case PerformReferenceMeasurement_msgid: {
            //_pi.request_loop_acquire_reference_sample();
            protocol_command_t cmd = PC_REQUEST_REFERENCE_SAMPLE_ACQUISITION;
            if (_pcQueue.try_add(cmd))
                log_message("acquire reference sample started");
        } break;
        case AutomaticOffsetAdjustmentRequest_msgid: {
            protocol_command_t cmd = PC_REQUEST_AUTOMATIC_OFFSET_ADJUSTMENT;
            if (_pcQueue.try_add(cmd))
                log_message("auto offset adjustment started");

            //_pi.request_automatic_offset_adjustment();
        } break;
        case NoiseFloorMeasurementRequest_msgid: {
            protocol_command_t cmd = PC_REQUEST_NOISE_FLOOR_MEASUREMENT;
            if (_pcQueue.try_add(cmd))
                log_message("noise floor measurement started");
        }
        break;
        case ThresholdConfigurationRequest_msgid: {
            send_threshold_configuration();
        } break;
        case ThresholdConfiguration_msgid: {
            ThresholdConfiguration tc = ThresholdConfiguration_init_default;
            if (pb_decode(&is, ThresholdConfiguration_fields, &tc)) {
                handle_threshold_configuration(tc);
            }
        } break;
        case SetSimulatedStateRequest_msgid: {
            SetSimulatedStateRequest r = SetSimulatedStateRequest_init_default;
            if (pb_decode(&is, SetSimulatedStateRequest_fields, &r)) {
                handle_simulated_state(r);
            }
        }
        break;
        case BoardRebootRequest_msgid: {
            log_message("executing reboot per PC request");
            watchdog_enable(1, false);
            sleep_ms(2);
        }
        break;
        case RestoreDefaultConfigurationRequest_msgid: {
            log_message("restoring default configuration per PC request");
            _params.clearParameters();
            watchdog_enable(1, false);
            sleep_ms(2);
        }
        break;
        default:
            log_message("unknown message received (id=%d)", msg->msgid);
            break;
    }
}

void ProtocolComm::log_message(const char *format, ...) {
    va_list argptr;
    va_start(argptr, format);
    static char logbuf[256];
    auto len = vsnprintf(logbuf, sizeof(logbuf), format, argptr);
    va_end(argptr);
    send_log_message(logbuf, len);
}

void ProtocolComm::copy_stats_to_measurement_sample(const SampleAnalyzer::sample_stats_t &stats, MeasurementSample &ms) {
    ms.pulse_len_us = stats.pulse_len_us;
    ms.mean_amp = stats.mean_amp;
    ms.mean_amp_filtered = stats.mean_amp_filtered;
    ms.min_amp = stats.min_amp;
    ms.max_amp = stats.max_amp;
    ms.lin_fit_valid = stats.lin_fit_valid;
    ms.lin_fit_offset = stats.lin_fit_offset;
    ms.lin_fit_slope = stats.lin_fit_slope;
    ms.exp_fit_valid = stats.exp_fit_valid;
    ms.exp_fit_exponent = stats.exp_fit_exponent;
    ms.exp_fit_intercept = stats.exp_fit_intercept;
    ms.weighted_delta_ref = stats.weighted_delta_ref;
    ms.integral = stats.integral;
}

void ProtocolComm::handle_sampling_configuration(SamplingConfiguration &cfg) {
    if (!_pi.set_sampling_configuration(cfg)) {
        log_message("failed to apply sampling configuration");
    } else {
        _params.parameters.md_sampling_configuration = cfg;
        _params.parameters.has_sampling_configuration = true;
        _params.writeParameters();
        _td.updateSamplingFrequency(_pi.get_sample_frequency_hz());
    }
}

void ProtocolComm::send_sampling_configuration() {
    auto sc = _pi.get_sampling_configuration();
    size_t framed_size=0;
    if (PARSER_FRAME_MESSAGE(&sc, SamplingConfiguration, _outmsgbuf,
                             sizeof(_outmsgbuf), &framed_size) == PARSER_RESULT_MSG_OK) {
        write_bytes(_outmsgbuf, framed_size);
    }
}

void ProtocolComm::send_threshold_configuration() {
    auto tc = _td.getConfiguration();
    size_t framed_size=0;
    if (PARSER_FRAME_MESSAGE(&tc, ThresholdConfiguration , _outmsgbuf,
                             sizeof(_outmsgbuf), &framed_size) == PARSER_RESULT_MSG_OK) {
        write_bytes(_outmsgbuf, framed_size);
    }
}

void ProtocolComm::handle_threshold_configuration(const ThresholdConfiguration &conf) {
    if (!_td.setConfiguration(conf)) {
        log_message("failed to apply threshold configuration!");
    } else {
        protocol_command_t cmd = PC_REQUEST_NOISE_FLOOR_MEASUREMENT;
        _pcQueue.try_add(cmd);

        _params.parameters.md_threshold_configuration = conf;
        _params.parameters.has_threshold_configuration = true;
        _params.writeParameters();
    }
}

void ProtocolComm::send_noise_floor_status_report() {
    auto tsr = _td.getStatusReport();
    size_t framed_size=0;
    if (PARSER_FRAME_MESSAGE(&tsr, ThresholdStatusReport , _outmsgbuf,
                             sizeof(_outmsgbuf), &framed_size) == PARSER_RESULT_MSG_OK) {
        write_bytes(_outmsgbuf, framed_size);
    }
}

void ProtocolComm::send_raw_sample(const RawSample &sample) {
    size_t framed_size=0;
    if (PARSER_FRAME_MESSAGE(&sample, RawSample, _outmsgbuf,
                             sizeof(_outmsgbuf), &framed_size) == PARSER_RESULT_MSG_OK) {
        write_bytes(_outmsgbuf, framed_size);
    }
}

void ProtocolComm::send_detection_measurement(const decay_sample_t &sample, SampleAnalyzer::sample_stats_t *stats,
                                              float *amplitude, float *phase) {

    const bool HIGH_RATE_STATS=false;
    auto freq = _pi.get_sample_frequency_hz();
    auto& undersample_proc = _pi.get_sampling_configuration().num_undersample_processing;
    if (undersample_proc != 0)
        freq = freq / undersample_proc;

    //this is rather crude calculation, as it does not take into account the number of samples actually
    //encoded (the struct size is maximum size, the actual size depends on number of samples analyzed)
    //and also serialization overhead
    unsigned long undersample_rate;
    if (HIGH_RATE_STATS)
        undersample_rate = 1;
    else
        undersample_rate = (freq * sizeof(DetectionMeasurement)) / DM_MAX_BANDWIDTH_BYTE_S;

    _numDetectionMeasurements++;
    if (_numDetectionMeasurements % undersample_rate == 0) {
        DetectionMeasurement dm = DetectionMeasurement_init_default;
        if (amplitude)
            dm.amplitude = *amplitude;
        if (phase)
            dm.phase = *phase;
        dm.timestamp_us = sample.timestamp_us;
        if (stats)
            copy_stats_to_measurement_sample(*stats, dm.sample);
        dm.has_sample = true;
        if (HIGH_RATE_STATS) {
            dm.sample.adc_sample_count = 0;
        } else {
            dm.sample.adc_sample_count = std::min(
                    (uint32_t) sizeof(dm.sample.adc_sample) / sizeof(dm.sample.adc_sample[0]),
                    (uint32_t) sample.num_samples);
            memcpy(dm.sample.adc_sample, sample.sample, dm.sample.adc_sample_count * sizeof(dm.sample.adc_sample[0]));
        }

        size_t framed_size = 0;
        if (PARSER_FRAME_MESSAGE(&dm, DetectionMeasurement, _outmsgbuf,
                                 sizeof(_outmsgbuf), &framed_size) == PARSER_RESULT_MSG_OK) {
            write_bytes(_outmsgbuf, framed_size);
        }
    }
}

void ProtocolComm::send_reference_measurement(const decay_sample_t &sample, SampleAnalyzer::sample_stats_t *stats) {
    ReferenceMeasurement rm = ReferenceMeasurement_init_default;
    rm.timestamp_us = time_us_64();
    if (stats)
        copy_stats_to_measurement_sample(*stats, rm.sample);
    rm.has_sample = true;
    rm.sample.adc_sample_count = std::min((uint32_t)sizeof(rm.sample.adc_sample)/sizeof(rm.sample.adc_sample[0]), (uint32_t)sample.num_samples);
    memcpy(rm.sample.adc_sample, sample.sample, rm.sample.adc_sample_count * sizeof(rm.sample.adc_sample[0]));

    size_t framed_size=0;
    if (PARSER_FRAME_MESSAGE(&rm, ReferenceMeasurement , _outmsgbuf,
                             sizeof(_outmsgbuf), &framed_size) == PARSER_RESULT_MSG_OK) {
        write_bytes(_outmsgbuf, framed_size);
    }
}

bool ProtocolComm::_initializeUART() {
    _uartInst = uart0;
    uint baudrate = 921600;
    uint tx_pin=0, rx_pin=1, espio02_pin=2;
    gpio_init(espio02_pin);
    gpio_set_dir(espio02_pin, false);
    uart_init(_uartInst, baudrate);

    // Set the TX and RX pins by using the function select on the GPIO
    // Set datasheet for more information on function select
    gpio_set_function(tx_pin, GPIO_FUNC_UART);
    gpio_set_function(rx_pin, GPIO_FUNC_UART);
    uart_set_hw_flow(_uartInst, false, false);

    // Set our data format
    uart_set_format(_uartInst, 8, 1, UART_PARITY_NONE);

    uart_set_fifo_enabled(_uartInst, true);
    int UART_IRQ = _uartInst == uart0 ? UART0_IRQ : UART1_IRQ;

    // And set up and enable the interrupt handlers
    irq_set_exclusive_handler(UART_IRQ, on_uart_rx);
    irq_set_enabled(UART_IRQ, true);
    // Now enable the UART to send interrupts - RX only
    uart_set_irq_enables(_uartInst, true, false); 
    _uartInitialized = true;
    return _uartInitialized;
}

void ProtocolComm::onUartRXIRQ() {
    while (uart_is_readable(_uartInst)) {
        _uartRXBuffer.try_add(uart_getc(_uartInst));
    }
}

void ProtocolComm::init() {
    _initializeUART();
}

bool ProtocolComm::send_system_status(const SystemStatus &status) {
    size_t framed_size=0;
    if (PARSER_FRAME_MESSAGE(&status, SystemStatus, _outmsgbuf,
                             sizeof(_outmsgbuf), &framed_size) == PARSER_RESULT_MSG_OK) {
        write_bytes(_outmsgbuf, framed_size);
        return true;
    }
    return false;
}

void ProtocolComm::handle_simulated_state(const SetSimulatedStateRequest &req) {
    if (req.has_battery_voltage) {
        _bm.set_simulated_battery_voltage(req.battery_voltage);
    } else {
        _bm.clear_simulated_battery_voltage();
    }

    if (req.has_vibro_pwm_dc) {
        _vibro.set_simulated_vibro_pwm_duty_cycle(req.vibro_pwm_dc);
    } else {
        _vibro.clear_simulated_vibro_pwm_duty_cycle();
    }
}
