# Firmware for RP2040 based pulse induction metal detector


## Flashing firmware using mass-storage bootloader

1. Plug in USB-C cable
2. Bridge J7 (always on power)
3. Bridge JP1 (boot in bootloader mode)
4. Mass storage device should appear on your computer. Copy pifw.uf2 to the device.

## Connection via WiFi-serial bridge

1. Flash [esp-link](https://gitlab.com/demineua/esp-link) to an ESP-01 module. The linked release is preconfigured for ESP01 and correct serial baud rate (921600).
2. Plug in the ESP01 module, start the board, connect to module's WiFi network.
3. Use ./main.py tcp:192.168.4.1:23 to start GUI.

# Development notes

## Organization of the code

Metal detector signal processing loop runs on core 0, audio output is handled by core 1 (audio generation is really jitter sensitive).

The code has recently been refactored in a slightly more manageable mess:
* Transmit pulse is handled by harware pwm. On pulse end, an interrupt is triggered, which starts PIO that reads ADC samples to memory via DMA.
* At the end of ADC read, DMA interrupt is triggered which pushes the read samples to `PISampler::decaySampleQueue`
* PIMD loop then gets those samples, calculates the statistics, and indicates the detection results via audio tone/LED bar 


## TODO
0. Make the detection usable 
2. fw: fix offset adjustment being different on startup vs later
1. GUI: display popup for errors reported by board
2. fw: implement MCU temperature measurement
3. fw: check if possible to correlate warmup of MCU and reference sample?
5. fw/gui: make automatic offset adjustment thresholds configurable
6. fw: detect overdamping before automatic offset adjustment

## Building firmware

Requirements: pico-sdk (developed against 1.5.1)

```bash
export PICO_SDK_PATH=~/opt/pico-sdk
mkdir build && cd build
cmake ..
make
```

By default, firmware is built for latest board (currently, PI_R2). To select another board, use PIFW_BOARD cmake argument (e.g. `cmake .. -DPIFW_BOARD=PI_R1`)

### Change audio sample language
To change audio sample language, set cmake argument `AUDIO_LANGUAGE` to `en`, `ua` (e.g. `cmake .. -DAUDIO_LANGUAGE=ua`)

## Flashing firmware

An easy way to flash firmware just with USB-C cable is:

1. Download/build [picotool](https://github.com/raspberrypi/picotool)
2. Plug in USB-C cable, press and hold the power button (it will need to be keep pressed for the whole duration of flashing process)
3. Run `picotool load pifw.uf2 -f`. This command uses vendor interface to reset the board in bootloader mode and flash the firmware.

 
## Setting up debugging in clion with JLink

1. Create a new debug configuration: Embedded GDB server
2. Download executable: always
3. 'target remote' args: ':2331'
4. GDB server: 'JLinkGDBServerCLExe'
5. GDB server args: '-select USB -device RP2040_M0_0 -endian little -if SWD -speed 1000 -noir -nogui -speed 4000 -noLocalhostOnly'
 
## Setting up debugging in clion with picoprobe
1. Create a new debug configuration: OpenOCD Download and run
2. Set board config file to rp2040.cfg (in this repo). The board config file selects programming interface to use (picoprobe), sets adapter speed to be sane (5MHz) and informs OpenOCD about it's target device (RP2040).

## Signal processing

The most basic measurement cycle consists of the following steps:

1. Apply a 50-200us long pulse.
2. Observe the decaying return signal: sample the amplified coil voltage via ADC for ~100us after the pulse.
3. Subtract or compare the sampled waveform with previously saved reference waveform (supposedly, without target present) 
   and produce an indication on the relative "strength" of the target to be output with LEDs/speaker.
4. Repeated the steps above at ~100-200Hz. 

Before acquisition of reference waveform, the offset of the amplified signal has to be adjusted by providing a 0..3.3V voltage reference (MD_OFFSET signal).

As with many things in life, the devil is in the details that were skipped in the simplified description:

1. Using pulses of varying length allow for finer distinction between ground and target.
2. The comparison step actually consists of computing a bunch of statistics on the acquired samples, the most important being: the rate of decay/slope of the signal, it's amplitude. 
 The comparison is then made based on the difference of the statistics of acquired sample and reference sample statistics.
3. It's important to select sensible threshold values for the comparisons of the statistics.
4. Results should be low-pass filtered for smooth output and less false positives due to noise (both external and internal).


For more details on the theory, please refer to chapters 11 and 12 of "Inside the Metal Detector" by George Overton, Carl Moreland.

## Actual signal waveforms

All the waveforms below are triggered at the falling edge of transmit pulse.

### Transmit pulse and coil return

Transmit pulse in blue (TX_MD), coil in yellow. As evident, pulse length is 120us for this example. Note the peak-to-peak coil voltage of 240V!

![transmit pulse](doc/gate-coil.png)

### Coil signal with and without amplification

Coil voltage in yellow, amplified signal (MDRX) in blue, with no target present. 
Note that the signal can be observed ~15us after the end of the pulse.

![amplified signal](doc/coil-amplified.png)

### Offset control
Amplified signal in blue, offset control signal in yellow (MD_OFFSET). As can be seen, the amplified signal decays to the level of offset.

![amplified signal vs offset](doc/amplified-offset.png)

### Target response

The following waveforms illustrate the effects on waveforms for several targets.

* Yellow trace: coil voltage
* Blue trace: amplified coil voltage
* White trace: reference amplified coil voltage (without target)

**Large metal target:**

![large target](doc/large-metal-target-response.png)

**Small metal target:**

![small target](doc/small-metal-target-response.png)

**0.2EUR coin at 15cm distance in air**

![coin target](doc/coin-15cm-response.png)

As can be seen, there is very little change in the signal for this target.


