//
// Created by R on 3/2/23.
//

#include "SampleAnalyzer.h"
#include <pico/stdlib.h>

#include "util/linalg.h"
using namespace linalg::aliases;

static StaticFIRFilter<float, 96> _meanAmpFIRFilter({-3.380000e-04f, -3.085172e-04f, -2.823400e-04f, -2.565110e-04f, -2.277302e-04f, -1.924029e-04f, -1.466946e-04f, -8.659135e-05f, -7.966145e-06f, 9.335139e-05f, 2.215018e-04f, 3.805208e-04f, 5.742639e-04f, 8.063309e-04f, 1.079993e-03f, 1.398123e-03f, 1.763127e-03f, 2.176889e-03f, 2.640711e-03f, 3.155271e-03f, 3.720583e-03f, 4.335967e-03f, 5.000034e-03f, 5.710674e-03f, 6.465058e-03f, 7.259653e-03f, 8.090244e-03f, 8.951969e-03f, 9.839361e-03f, 1.074641e-02f, 1.166660e-02f, 1.259304e-02f, 1.351846e-02f, 1.443538e-02f, 1.533612e-02f, 1.621295e-02f, 1.705819e-02f, 1.786426e-02f, 1.862380e-02f, 1.932977e-02f, 1.997557e-02f, 2.055504e-02f, 2.106265e-02f, 2.149346e-02f, 2.184331e-02f, 2.210875e-02f, 2.228717e-02f, 2.237683e-02f, 2.237683e-02f, 2.228717e-02f, 2.210875e-02f, 2.184331e-02f, 2.149346e-02f, 2.106265e-02f, 2.055504e-02f, 1.997557e-02f, 1.932977e-02f, 1.862380e-02f, 1.786426e-02f, 1.705819e-02f, 1.621295e-02f, 1.533612e-02f, 1.443538e-02f, 1.351846e-02f, 1.259304e-02f, 1.166660e-02f, 1.074641e-02f, 9.839361e-03f, 8.951969e-03f, 8.090244e-03f, 7.259653e-03f, 6.465058e-03f, 5.710674e-03f, 5.000034e-03f, 4.335967e-03f, 3.720583e-03f, 3.155271e-03f, 2.640711e-03f, 2.176889e-03f, 1.763127e-03f, 1.398123e-03f, 1.079993e-03f, 8.063309e-04f, 5.742639e-04f, 3.805208e-04f, 2.215018e-04f, 9.335139e-05f, -7.966145e-06f, -8.659135e-05f, -1.466946e-04f, -1.924029e-04f, -2.277302e-04f, -2.565110e-04f, -2.823400e-04f, -3.085172e-04f, -3.380000e-04f});

SampleAnalyzer::SampleAnalyzer() {
}

bool
SampleAnalyzer::sample_compute_linear_slope(uint32_t num_samples, const float *sample_buf, float &slope, float &offset) {
    if (num_samples < 2)
        return false;
    float xsumsq = 0.0f;
    float xsum = 0.0f;
    float xysum = 0.0f;
    float ysum = 0.0f;
    for (int i=0; i < num_samples; i++) {
        float x = i * US_PER_SAMPLE;
        float y = sample_buf[i];
        xsumsq += x * x;
        xsum += x;
        xysum += x * y;
        ysum += y;
    }

    float2x2 matA {{xsumsq, xsum}, {xsum, (float)num_samples}};
    float2x1 matB{{xysum, ysum}};

    auto matAInv = linalg::inverse(matA);
    auto matx = linalg::mul(matAInv, matB);
    slope = matx[0][0];
    offset = matx[0][1];

    return true;
}

bool
SampleAnalyzer::sample_compute_exponential_decay_slope(uint32_t num_samples, const float *sample_buf, float &exponent,
                                                       float &intercept) {
    if (num_samples < 2)
        return false;
    float xsqysum = 0.0f;
    float xysum = 0.0f;
    float ysum = 0.0f;
    float ylogysum = 0.0f;
    float xylogysum = 0.0f;
    float maxy = 0.0f;
    for (int i=0; i < num_samples; i++) {
        float y = sample_buf[i];
        if (y > maxy)
            maxy = y;
    }
    for (int i=0; i < num_samples; i++) {
        float x = i * US_PER_SAMPLE;
        float y = maxy - sample_buf[i] + 0.000001f;
        xsqysum += x*x*y;
        xysum += x * y;
        ysum += y;
        auto logy = logf(y);
        ylogysum += y * logy;
        xylogysum += x * y * logy;
    }

    float2x2 matA {{ysum, xysum}, {xysum, xsqysum}};
    float2x1 matB{{ylogysum, xylogysum}};
    auto matAInv = linalg::inverse(matA);
    auto matx = linalg::mul(matAInv, matB);
    intercept= expf(matx[0][0]);
    exponent = matx[0][1];

    return true;
}

float SampleAnalyzer::sample_compute_weighted_delta_ref(uint32_t num_samples, const float *sample_buf,
                                                        const float *ref_sample_buf) {
    float weighted_delta_ref = 0.0f;
    for (uint32_t i=0; i < num_samples; i++) {
        float weight = 1.0f - (1.0f / num_samples * i);
        weight = weight*weight;
        weighted_delta_ref += (sample_buf[i] - ref_sample_buf[i]) * weight;
    }
    return weighted_delta_ref;
}

SampleAnalyzer::sample_stats_t SampleAnalyzer::sample_compute_stats(const decay_sample_t &sample, const decay_sample_t &ref_sample,
                                                    uint32_t num_samples_to_analyze,
                                                    bool compute_expensive_stats) {
    auto stats = sample_compute_stats(sample, num_samples_to_analyze, compute_expensive_stats);
    stats.weighted_delta_ref = sample_compute_weighted_delta_ref(num_samples_to_analyze, sample.sample, ref_sample.sample);
    return stats;
}

SampleAnalyzer::sample_stats_t
SampleAnalyzer::sample_compute_stats(const decay_sample_t &sample, uint32_t num_samples_to_analyze,
                                     bool compute_expensive_stats) {
    assert(num_samples_to_analyze > 0 && num_samples_to_analyze <= sample.num_samples);
    sample_stats_t stats;
    stats.min_amp = 255.0f;
    stats.max_amp = 0.0f;
    stats.clip_end_idx = 0;
    float sum=0.0f;
    uint32_t count = 0;
    uint32_t top_clipping_count = 0;
    for (int i=0; i < num_samples_to_analyze; i++) {
        auto s = sample.sample[i];
        if (s < stats.min_amp)
            stats.min_amp = s;
        if (s > stats.max_amp)
            stats.max_amp = s;
        if (s >= CLIPPING_TOP_THRESHOLD)
            top_clipping_count++;

        sum += s;
        count++;

        if (s < 254 && stats.clip_end_idx == 0) {
            stats.clip_end_idx = i;
        }
    }
    stats.top_clipping_detected = top_clipping_count >= (int)(num_samples_to_analyze*CLIPPING_COUNT_THRESHOLD_PCT);

    stats.integral = sum;
    if (count != 0)
        stats.mean_amp = sum / count;
    else
        stats.mean_amp = 0.0f;

    auto maf = _meanAmpFIRFilter.apply_filter(stats.mean_amp);
    if (maf)
        stats.mean_amp_filtered = maf.value();
    else
        stats.mean_amp_filtered = stats.mean_amp;

    if (compute_expensive_stats) {
        stats.lin_fit_valid = sample_compute_linear_slope(
                num_samples_to_analyze, sample.sample, stats.lin_fit_slope, stats.lin_fit_offset);
        stats.exp_fit_valid = sample_compute_exponential_decay_slope(
                num_samples_to_analyze, sample.sample, stats.exp_fit_exponent, stats.exp_fit_intercept
        );
    } else {
        stats.lin_fit_offset = stats.lin_fit_slope = stats.exp_fit_exponent = stats.exp_fit_intercept = 0.0f;
        stats.lin_fit_valid = stats.exp_fit_valid = false;
    }

    stats.weighted_delta_ref = 0;
    return stats;
}
