//
// Created by R on 6/26/23.
//

#include "ThresholdDetectorSM.h"
#include <algorithm>
#include <cmath>
#include "util/linalg.h"
using namespace linalg::aliases;


ThresholdDetectorSM::ThresholdDetectorSM():
        _initialized(false),
        _detectionMode(DETECTION_MODE_DYNAMIC),
        _samplingFrequencyHz(0)
{
    _conf.detectionStatistic = ThresholdConfiguration_DetectionStatistic_MEAN_AMPLITUDE_FILTERED;
    _conf.sensitivity = 1.0f;
    _conf.noiseFloorMultiplier = 1.3f;
    _conf.minNoiseFloor = 10.0f;
    _conf.compute_expensive_stats = false;
    _conf.dynamicModeAlpha = 5.0f;
    resetState();
}

void ThresholdDetectorSM::resetState() {
    _initialized = false;
    _meanAmpNF.count = 0;
    memset(&_dmState, 0, sizeof(_dmState));
    _dynRefLPF.setOutput(0.0f);
}

void ThresholdDetectorSM::setSensitivity(const float &sensitivity) {
    _conf.sensitivity = std::clamp(sensitivity, 0.0f, 1.0f);
}

bool ThresholdDetectorSM::setConfiguration(const ThresholdConfiguration &conf) {
    _conf = conf;
    _conf.sensitivity = std::clamp(_conf.sensitivity, 0.0f, 1.0f);
    _conf.noiseFloorMultiplier = std::clamp(_conf.noiseFloorMultiplier, 0.0f, 10.0f);
    updateDynamicModeLPF();
    return true;
}

ThresholdStatusReport ThresholdDetectorSM::getStatusReport() const {
    ThresholdStatusReport sr;
    sr.noiseFloor = _meanAmpNF.noise;
    sr.meanValue = _meanAmpNF.mean;
    sr.minValue = _meanAmpNF.minValue;
    sr.meanValueSlope = _meanAmpNF.slope;
    sr.measuredNoiseFloor = _meanAmpNF.measuredNoiseFloor;
    return sr;
}

float ThresholdDetectorSM::setSensitivityFromUserSetting(int8_t user_sens) {
    const float sens_levels[] = {0.825f, 0.85f, 0.875f, 0.9f, 0.925f, 0.95f, 0.975f, 0.99f, 0.999f, 1.0f};
    if (user_sens < 1) {
        user_sens = 1;
    } else if (user_sens > 10) {
        user_sens = 10;
    }
    _conf.sensitivity = sens_levels[user_sens - 1];
    _conf.normalizedSensitivity = user_sens * 0.1f;
    return _conf.sensitivity;
}

bool ThresholdDetectorSM::consumeSample(const decay_sample_t &sample, const SampleAnalyzer::sample_stats_t &stats,
                                        const decay_sample_t &refSample, const SampleAnalyzer::sample_stats_t &refStats,
                                        float &result) {
    result = 0.0f;
    bool statValid = true;
    float Vref; //reference stat value
    float Vact; //current stat value
    switch (_conf.detectionStatistic) {
        case ThresholdConfiguration_DetectionStatistic_MEAN_AMPLITUDE:
            Vref = refStats.mean_amp;
            Vact = stats.mean_amp;
            break;
        case ThresholdConfiguration_DetectionStatistic_MEAN_AMPLITUDE_FILTERED:
            Vref = refStats.mean_amp_filtered;
            Vact = stats.mean_amp_filtered;
            break;
        case ThresholdConfiguration_DetectionStatistic_EXPONENTIAL_FIT_EXPONENT:
            Vref = refStats.exp_fit_exponent;
            Vact = stats.exp_fit_exponent;
            statValid = stats.exp_fit_valid && refStats.exp_fit_valid;
            break;
        case ThresholdConfiguration_DetectionStatistic_EXPONENTIAL_FIT_INTERCEPT:
            Vref = refStats.exp_fit_intercept;
            Vact = stats.exp_fit_intercept;
            statValid = stats.exp_fit_valid && refStats.exp_fit_valid;
            break;
        case ThresholdConfiguration_DetectionStatistic_INTEGRAL:
            Vref = refStats.integral;
            Vact = stats.integral;
            break;
        case ThresholdConfiguration_DetectionStatistic_LINEAR_FIT_OFFSET:
            Vref = refStats.lin_fit_offset;
            Vact = stats.lin_fit_offset;
            statValid = stats.lin_fit_valid && refStats.lin_fit_valid;
            break;
        case ThresholdConfiguration_DetectionStatistic_LINEAR_FIT_SLOPE:
            Vref = refStats.lin_fit_slope;
            Vact = stats.lin_fit_slope;
            statValid = stats.lin_fit_valid && refStats.lin_fit_valid;
            break;
        case ThresholdConfiguration_DetectionStatistic_WEIGHTED_DELTA:
            Vref = 0.0f;
            Vact = stats.weighted_delta_ref;
            break;
        default:
            Vref = refStats.mean_amp;
            Vact = stats.mean_amp;
            break;
    }

    if (!statValid)
        return false;

    auto time_us = time_us_64();
    if (_initialized) {

        float noiseSens = _meanAmpNF.noise + _meanAmpNF.noise * 100.0f * (1.0f - _conf.sensitivity);
        if (_detectionMode == DETECTION_MODE_STATIC) {
            float absDelta = fabsf(Vref - Vact);
            if (absDelta > noiseSens) {
                float range = fabsf(Vref) - noiseSens;
                float ampNorm = fabsf((range - Vact) / range);
                float tmpResult = ampNorm;
                result = std::clamp(tmpResult, 0.0f, 1.0f);
            }
            return true;
        } else if (_detectionMode == DETECTION_MODE_DYNAMIC) {
             if (_dmState.last_sample_timestamp_us == 0) {
                _dmState.VactLPF = Vact;
                _dynRefLPF.setOutput(Vact);
             } else {
					
                 //float absDelta = fabsf(_dmState.VactLPF - Vact);

                 float range = (1.0f - _conf.normalizedSensitivity) * 3000.0f + 100.0f; // roughly 100..1000 counts, depending on sensitivity

                 float ampNorm = fabsf(std::clamp(_dmState.VactLPF - Vact, 0.0f, range) / range); // maps the difference between current signal and LPF to 0..1
                 float tmpResult = std::log10(4 * ampNorm + 1) / std::log10(2);  //logarithmic adjustment- change in values near zero result in more change of the output
                // Inline custom curve mapping
				/*  float tmpResult;
					{
						// Define breakpoints and corresponding values
						float x[] = {0.00, 0.01, 0.05, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 1.0};
						float y[] = {0, 0.01, 3, 6, 6, 7, 7, 7, 7, 7, 7, 6}; // Desired output values

						// Find the interval containing ampNorm and perform linear interpolation
						for (int i = 0; i < 10; i++) {
							if (ampNorm >= x[i] && ampNorm <= x[i + 1]) {
								float t = (ampNorm - x[i]) / (x[i + 1] - x[i]);
								tmpResult = (y[i] + t * (y[i + 1] - y[i])) / 8.0f; // Normalize to [0..1]
								break;
							}
						}
					} */
                 result = std::clamp(tmpResult, 0.01f, 1.0f);
                 _dmState.VactLPF = _dynRefLPF.update(Vact);
             }
             _dmState.last_sample_timestamp_us = time_us;
            return true;
        }
    } else {
        _meanAmpNF.sampleMeanHistory[_meanAmpNF.count] = Vact;
        _meanAmpNF.count++;

        if (_meanAmpNF.count >= NOISE_FLOOR_NUM_SAMPLES) {
            // perform initialization- calculate noise floor and mean
            float slope, offset;
            computeSampleMeanHistoryLinearFit(slope, offset);
            //correct for the increasing/decreasing slope of the measurements during noise floor measurement
            //this is particularly important for the turn-on phase, where the measured mean value is increasing
            //due to temperature
            for (uint32_t i=0; i < NOISE_FLOOR_NUM_SAMPLES; i++) {
                _meanAmpNF.sampleMeanHistory[i] -= slope * i;
            }
            float sum=0.0f;
            float minValue = 1e6f;
            float maxValue =  -1e6f;
            unsigned int count = 0;
            for (auto& s: _meanAmpNF.sampleMeanHistory) {
                if (s > maxValue)
                    maxValue = s;
                if (s < minValue)
                    minValue = s;
                sum += s;
                count++;
            }
            _meanAmpNF.mean = sum / count;
            _meanAmpNF.measuredNoiseFloor = (_meanAmpNF.mean - minValue) * _conf.noiseFloorMultiplier;
            _meanAmpNF.noise = std::clamp(_meanAmpNF.measuredNoiseFloor, _conf.minNoiseFloor, 16384.0f);
            _meanAmpNF.slope = slope;
            _meanAmpNF.minValue = minValue;
            _initialized = true;
        }
        return false;
    }
    return false;
}

void ThresholdDetectorSM::computeSampleMeanHistoryLinearFit(float &slope, float &offset) {
    //performs a linear fit on _meanAmpNF.sampleMeanHistory entries
    //returned slope can be used to exclude linear rise/fall of the sample trend (e.g. rising temperature effects)
    float xsumsq = 0.0f;
    float xsum = 0.0f;
    float xysum = 0.0f;
    float ysum = 0.0f;
    for (int i=0; i < NOISE_FLOOR_NUM_SAMPLES; i++) {
        float x = i;
        float y = _meanAmpNF.sampleMeanHistory[i];
        xsumsq += x * x;
        xsum += x;
        xysum += x * y;
        ysum += y;
    }

    float2x2 matA {{xsumsq, xsum}, {xsum, (float)NOISE_FLOOR_NUM_SAMPLES}};
    float2x1 matB{{xysum, ysum}};

    auto matAInv = linalg::inverse(matA);
    auto matx = linalg::mul(matAInv, matB);
    slope = matx[0][0];
    offset = matx[0][1];

}

void ThresholdDetectorSM::updateSamplingFrequency(uint32_t freq) {
    _samplingFrequencyHz = freq;
    updateDynamicModeLPF();
}

void ThresholdDetectorSM::updateDynamicModeLPF() {
    if (_samplingFrequencyHz > 0)
        _dynRefLPF.reconfigureFilter(1.0f / _samplingFrequencyHz, _conf.dynamicModeAlpha);
}
