//
// Created by R on 6/26/23.
//

#ifndef PIFW_THRESHOLDDETECTORSM_H
#define PIFW_THRESHOLDDETECTORSM_H

/**
 * Threshold detector is responsible for processing the sample stats and outputting "metal detected" signal.
 *
 * The output range is 0 .. 1.0f. Input range is dynamically determined during initialization, where a number of
 * samples are fed to the detector which measures background noise, thus determining minimum detection threshold.
 *
 * Sensitivity of the detector can be adjusted so that it can ignore small targets. Sensitivity range is 0 .. 1.0f, where
 * 1.0f is the maximum sensitivity (i.e. it reacts to the smallest possible target).
 *
 */
#include "SampleAnalyzer.h"
#include "sampler/PISampler.h"
#include "LowPassFilter.hpp"
#include "structures.h"

class ThresholdDetectorSM {
public:
    ThresholdDetectorSM();

    bool needsInitialization() const { return !_initialized; }

    void resetState();

    void setSensitivity(const float& sensitivity);

    /**
     * Performs detection and provides detection result if noise floor measurement is finished
     * @param sample
     * @param stats
     * @param result
     * @return
     */
    bool consumeSample(const decay_sample_t& sample, const SampleAnalyzer::sample_stats_t& stats,
                       const decay_sample_t& refSample, const SampleAnalyzer::sample_stats_t& refStats,
                       float& result);

    ThresholdConfiguration getConfiguration() const { return _conf; }

    bool setConfiguration(const ThresholdConfiguration& conf);

    void updateSamplingFrequency(uint32_t freq);

    ThresholdStatusReport getStatusReport() const;

    /**
     * Sets sensitivity based on user setting 1..10
     * @param user_sens
     * @return
     */
    float setSensitivityFromUserSetting(int8_t user_sens);

    void setDetectionModeFromUserSetting(detection_mode_t mode) { _detectionMode = mode; }

    float getDynamicModeVActLPF() const {
         return _dmState.VactLPF;
    }
protected:
    void updateDynamicModeLPF();
    bool _initialized;

    ThresholdConfiguration _conf;
    static const unsigned int NOISE_FLOOR_NUM_SAMPLES=1600;

    struct {
        float mean;  // mean value during noise floor measurement
        float noise; // noise = max. measured difference from mean

        unsigned int count;
        float sampleMeanHistory[NOISE_FLOOR_NUM_SAMPLES];
        float slope;
        float minValue;
        float measuredNoiseFloor;
    } _meanAmpNF;

    struct {
        uint64_t last_sample_timestamp_us;
        float VactLPF;
    } _dmState; //dynamic mode state

    LowPassFilter    _dynRefLPF;
    detection_mode_t _detectionMode;
    uint32_t         _samplingFrequencyHz;
    void computeSampleMeanHistoryLinearFit(float& slope, float& offset);
};


#endif //PIFW_THRESHOLDDETECTORSM_H
