//
// Created by R on 23.18.2.
//

#include "UI.h"
#include <hardware/pwm.h>
#include <hardware/clocks.h>
#include <stdio.h>
#include <cstring>
#include <cmath>

#include "ws2812.pio.h"
#include <hardware/dma.h>
#include <pico/bootrom.h>
#include "board.h"

bool button_timer_callback(repeating_timer_t* t) {
    assert(t->user_data);
    ((UI*)t->user_data)->button_handler();
    return true;
}


//important: order of pins must be the same as button_t
const uint32_t UI::PIN_BUTTON[BUTTON_END] = {
        8, //power/back
        15,//up
        16,//down
        17,//ok/menu
};

const UI::button_polarity_t UI::PIN_BUTTON_POLARITY[BUTTON_END] = {
        UI_PWR_BUTTON_POLARITY,
        UI_UP_BUTTON_POLARITY,
        UI_DOWN_BUTTON_POLARITY,
       UI_MENU_BUTTON_POLARITY
};

bool UI::init() {
    _init_gpio();
    _init_pio();
    _pixel_dma_chan = dma_claim_unused_channel(true);
    auto c = dma_channel_get_default_config(_pixel_dma_chan);
    channel_config_set_transfer_data_size(&c, DMA_SIZE_32);
    channel_config_set_read_increment(&c, true);
    channel_config_set_write_increment(&c, false);
    channel_config_set_dreq(&c, pio_get_dreq(_pio_ws2812, _pio_ws2812_sm, true));
    dma_channel_configure(_pixel_dma_chan, &c, &(_pio_ws2812->txf[_pio_ws2812_sm]), _pixel_dma_buffer, num_ws2812_active_pixels, false);

    add_repeating_timer_ms(1000 / BUTTON_HANDLER_FREQ_HZ, &button_timer_callback, (void*)this, &_timer_buttons);

    return true;
}

UI::UI(HAL &hal) :
    _in_menu(false), _selected_menu_item(MENU_NONE), _current_sensitivity(10), _current_volume(1),
    _current_backlight_brightness(5), _backlight_enabled(true),
    _hal(hal),
    _audio(hal.audio), _poweringOff(false), _detectionMode(DETECTION_MODE_STATIC), _startupDone(false),
    _startupButtonInhibitTs(0),
    _battery_soc(0),
    _vibro(hal.vibro),
    _current_vibro_power(1),
    _potentialShutdownCommand(false)
{
    memset(_button_prev_state, 0, sizeof(_button_prev_state));
    memset(_button_prev_state_ts, 0, sizeof(_button_prev_state_ts));
    memset(_pixel_buffer, 0, sizeof(_pixel_buffer));
    memset(_pixel_dma_buffer, 0, sizeof(_pixel_dma_buffer));
    num_ws2812_active_pixels = BOARD_HAS_BACKLIGHT ? 14 : 10; // last two pixels are backlight, if the board has it
}

void UI::_init_gpio() {
    gpio_init(PIN_POWER_EN);
    gpio_set_dir(PIN_POWER_EN, true);
    gpio_put(PIN_POWER_EN, true);

    for (int i=0; i < BUTTON_END; i++) {
        auto gpio = PIN_BUTTON[i];
        auto pol = PIN_BUTTON_POLARITY[i];
        gpio_init(gpio);
        gpio_set_dir(gpio, false);
        if (pol == POL_ACTIVE_HIGH)
            gpio_pull_down(gpio);
        else {
            if (gpio != 16 || (gpio == 16 && !BOARD_DISABLE_PULLUP_GPIO16))
                gpio_pull_up(gpio);
        }
    }
}

bool UI::get_button_state(UI::button_t btn) {
    assert(btn < BUTTON_END);
    auto pol = PIN_BUTTON_POLARITY[btn];
    auto state = gpio_get(PIN_BUTTON[btn]);
    return (PIN_BUTTON_POLARITY[btn] == POL_ACTIVE_HIGH) == state;
}

void UI::_init_pio() {
    _pio_ws2812 = pio1;
    _pio_ws2812_sm = pio_claim_unused_sm(_pio_ws2812, true);
    auto offset = pio_add_program(_pio_ws2812, &ws2812_program);
    ws2812_program_init(_pio_ws2812, _pio_ws2812_sm, offset, PIN_WS2812, 800000, false);
}

void UI::set_pixel_data(uint8_t pixel, uint8_t r, uint8_t g, uint8_t b) {
    assert(pixel <  NUM_WS2812_PIXELS);
    _pixel_buffer[pixel] =  {r, g, b};
}

void UI::transmit_pixel_data() {
    for (int i=0; i < NUM_WS2812_PIXELS; i++)
        _pixel_dma_buffer[i] = urgb_u32(_pixel_buffer[i].r, _pixel_buffer[i].g, _pixel_buffer[i].b) << 8u;
    dma_channel_set_read_addr(_pixel_dma_chan, _pixel_dma_buffer, true);
}


void UI::led_animation_blocking() {
    int num_pixels = 10;
    for (int i=0;i<num_pixels;i++) {
        set_pixel_data(i, 0, 0, 0);
    }
    transmit_pixel_data();
    sleep_ms(1);
    //bottom->top
    for (int j=0;j<num_pixels;j++) {
        for (int i = 0; i < num_pixels; i++) {
            uint8_t r=0, g=0, b=0;
            if (i <= j)
                fillColormapPixelValue(i, COLOR_MAP_GREEN_TO_RED, r, g, b);

            set_pixel_data(i, r, g, b);
        }
        transmit_pixel_data();
        sleep_ms(20);
    }
    sleep_ms(20);
    for (int j=0;j<num_pixels;j++) {
        for (int i = 0; i < num_pixels; i++) {
            uint8_t r=0, g=0, b=0;
            fillColormapPixelValue(i, COLOR_MAP_GREEN_TO_RED, r, g, b);
            if (i <= j) {
                r = 0;
                g = 0;
                b = 0;
            }

            set_pixel_data(i, r, g, b);
        }
        transmit_pixel_data();
        sleep_ms(20);
    }
    sleep_ms(20);

    //top->bottom
    for (int j=num_pixels-1;j>=0;j--) {
        for (int i = 0; i < num_pixels; i++) {
            uint8_t r=0, g=0, b=0;
            fillColormapPixelValue(i, COLOR_MAP_GREEN_TO_RED, r, g, b);
            if (i <= j) {
                r = 0;
                g = 0;
                b = 0;
            }

            set_pixel_data(i, r, g, b);
        }

        transmit_pixel_data();
        sleep_ms(20);
    }
    for (int j=num_pixels-1;j>=0;j--) {
        for (int i = 0; i < num_pixels; i++) {
            uint8_t r=0, g=0, b=0;
            if (i <= j)
                fillColormapPixelValue(i, COLOR_MAP_GREEN_TO_RED, r, g, b);
            set_pixel_data(i, r, g, b);
        }

        transmit_pixel_data();
        sleep_ms(20);
    }


    for (int i=0;i<num_pixels;i++) {
        set_pixel_data(i, 0, 0, 0);
    }

    transmit_pixel_data();
    sleep_ms(1);

    /*
    sleep_ms(100);
    for (int i=num_pixels;i>0;i--) {
        put_pixel(0, 0, 0);
    }
     */
}

void UI::button_printf_state() {
    printf("Button state    POWER: %d, UP: %d, DOWN: %d, OK: %d\n",
           get_button_state(BUTTON_POWER_BACK),
           get_button_state(BUTTON_UP),
           get_button_state(BUTTON_DOWN),
           get_button_state(BUTTON_OK_MENU)
           );
}

void UI::button_handler() {
    auto time_us = time_us_64();

    //ignore button press right after startup, as it can be the same as power on
    if (!_startupDone || time_us < _startupButtonInhibitTs)
        return;

    for (int b=0;b<BUTTON_END;b++) {
        auto state = get_button_state((button_t)b);
        auto prevstate = _button_prev_state[b];
        auto dt_ms = (time_us - _button_prev_state_ts[b]) / 1000;

        if (state && !prevstate) {
            // key pressed
        }

        if (state && prevstate) {
            //turn-off animation on the LEDs
            if (b == BUTTON_POWER_BACK) {
                if (dt_ms > POWER_OFF_INDICATION_START_MS) {
                    _potentialShutdownCommand = true;
                    const uint32_t ms_total = POWER_OFF_PRESS_DURATION_MS - POWER_OFF_INDICATION_START_MS;
                    uint32_t       ms_remaining = POWER_OFF_PRESS_DURATION_MS - dt_ms;
                    uint8_t level = ms_remaining * 10 / ms_total + 1;
                    set_led_bar_level(level, COLOR_MAP_WHITE);
                }
            }
            // keeps pressed
            if (dt_ms > POWER_OFF_PRESS_DURATION_MS) {
                //long press
                if (b == BUTTON_POWER_BACK) {
                    send_button_event(EVENT_POWER_OFF);
                }
            }
        }

        if (!state && prevstate) {
            if (_potentialShutdownCommand && b == BUTTON_POWER_BACK) {
                _potentialShutdownCommand = false;
                set_led_bar_level(0);
            }
            // key released
            if (dt_ms > 3000) {
                //long press released

            } else {
                switch (b) {
                    case BUTTON_UP:
                        send_button_event(EVENT_UP);
                        break;
                    case BUTTON_DOWN:
                        send_button_event(EVENT_DOWN);
                        break;
                    case BUTTON_OK_MENU:
                        send_button_event(EVENT_ENTER_MENU);
                        break;
                    case BUTTON_POWER_BACK:
                        send_button_event(EVENT_BACK);
                        break;
                    default:
                        break;
                }
            }
        }
        if (state != prevstate) {
            _button_prev_state_ts[b] = time_us;
        }
        _button_prev_state[b] = state;
    }
    if (_selected_menu_item != MENU_NONE) {

        button_t timeout_buttons[] = {BUTTON_UP, BUTTON_DOWN, BUTTON_OK_MENU};
        bool timeout = true;
        for (auto b: timeout_buttons) {
            if (_button_prev_state_ts[b] != 0 && (time_us - _button_prev_state_ts[b]) < MENU_TIMEOUT_US) {
                timeout = false;
            }
        }

        if (timeout)
            send_button_event(EVENT_EXIT_MENU);
    }

    switch (_selected_menu_item) {
        case MENU_SENSITIVITY:
            set_led_bar_level(_current_sensitivity);
            break;
        case MENU_VOLUME:
            set_led_bar_level(_current_volume);
            break;
        case MENU_VIBRO_POWER:
            set_led_bar_level(_current_vibro_power);
            break;
        case MENU_BRIGHTNESS:
            set_led_bar_level(_current_backlight_brightness);
            break;
        case MENU_BACKLIGHT:
            if (_backlight_enabled)
                set_led_bar_level(10);
            else
                set_led_bar_level(0);
            break;
        case MENU_DETECTION_MODE:
            if (_detectionMode == DETECTION_MODE_STATIC)
                set_led_bar_level(10);
            else
                set_led_bar_level(0);
    }
}

bool UI::send_button_event(UI::event_t evt) {
    return _button_event_queue.try_add(evt);
}

void UI::process_events() {
    event_t evt;
    if (_button_event_queue.try_remove(evt)) {
        switch (evt) {
            case EVENT_POWER_OFF:
                if (!_poweringOff) {
                    gpio_put(PIN_POWER_EN, false);
                    push_ui_command(CMD_INDICATE_POWER_OFF);
                    _poweringOff = true;
                }
            break;
            case EVENT_UP:
                switch (_selected_menu_item) {
                    case MENU_SENSITIVITY:
                        inc_setting(_current_sensitivity);
                        break;
                    case MENU_VOLUME:
                        inc_setting(_current_volume);
                        push_ui_command(CMD_INDICATE_VOLUME_CHANGE);
                        break;
                    case MENU_VIBRO_POWER:
                        inc_setting(_current_vibro_power);
                        push_ui_command(CMD_INDICATE_VIBRO_POWER_CHANGE);
                        break;
                    case MENU_BRIGHTNESS:
                        inc_setting(_current_backlight_brightness);
                        break;
                    case MENU_BACKLIGHT:
                        _backlight_enabled = true;
                        _audio.playSystemSound(PWMAudio::SOUND_ON);
                        break;
                    case MENU_DETECTION_MODE:
                        _detectionMode = DETECTION_MODE_STATIC;
                        _audio.playSystemSound(PWMAudio::SOUND_STATIC);
                        break;
                }
                break;
            case EVENT_DOWN:
                switch (_selected_menu_item) {
                    case MENU_SENSITIVITY:
                        dec_setting(_current_sensitivity);
                        break;
                    case MENU_VOLUME:
                        dec_setting(_current_volume);
                        push_ui_command(CMD_INDICATE_VOLUME_CHANGE);
                        break;
                    case MENU_VIBRO_POWER:
                        dec_setting(_current_vibro_power, true);
                        push_ui_command(CMD_INDICATE_VIBRO_POWER_CHANGE);
                        break;
                    case MENU_BRIGHTNESS:
                        dec_setting(_current_backlight_brightness);
                        break;
                    case MENU_BACKLIGHT:
                        _backlight_enabled = false;
                        _audio.playSystemSound(PWMAudio::SOUND_OFF);
                        break;
                    case MENU_DETECTION_MODE:
                        _detectionMode = DETECTION_MODE_DYNAMIC;
                        _audio.playSystemSound(PWMAudio::SOUND_DYNAMIC);
                        break;
                    case MENU_NONE:
                        push_ui_command(CMD_INDICATE_BATTERY_LEVEL);
                        break;
                }
                break;
            case EVENT_ENTER_MENU:
                switch (_selected_menu_item) {
                    case MENU_NONE:
                        push_ui_command(CMD_ENTER_MENU);
                        _selected_menu_item = MENU_SENSITIVITY;
                        _audio.playSystemSound(PWMAudio::SOUND_SENSITIVITY_ADJUSTMENT);
                        break;
                    case MENU_SENSITIVITY:
                        _selected_menu_item = MENU_VOLUME;
                        _audio.playSystemSound(PWMAudio::SOUND_VOLUME_ADJUSTMENT);
                        break;
                    case MENU_VOLUME:
                        _selected_menu_item = MENU_VIBRO_POWER;
                        _audio.playSystemSound(PWMAudio::SOUND_VIBRATION_POWER_ADJUSTMENT);
                        break;
                    case MENU_VIBRO_POWER:
                        _selected_menu_item = MENU_DETECTION_MODE;
                        _audio.playSystemSound(PWMAudio::SOUND_DETECTION_MODE);
                        break;
                    case MENU_DETECTION_MODE:
                        _selected_menu_item = MENU_BRIGHTNESS;
                        _audio.playSystemSound(PWMAudio::SOUND_BRIGHTNESS_ADJUSTMENT);
                        break;
                    case MENU_BRIGHTNESS:
                        if (BOARD_HAS_BACKLIGHT) {
                            _selected_menu_item = MENU_BACKLIGHT;
                            _audio.playSystemSound(PWMAudio::SOUND_BACKLIGHT);
                        } else {
                            _selected_menu_item = MENU_NONE;
                            send_button_event(EVENT_EXIT_MENU);
                        }
                        break;
                    case MENU_BACKLIGHT:
                        _selected_menu_item = MENU_NONE;
                        send_button_event(EVENT_EXIT_MENU);
                        break;
                }
                break;
            case EVENT_EXIT_MENU:
                _selected_menu_item = MENU_NONE;
                push_ui_command(CMD_UPDATE_SETTINGS);
                push_ui_command(CMD_EXIT_MENU);
                break;
            case EVENT_BACK:
                if (_selected_menu_item == MENU_NONE) {
                    if (_detectionMode == DETECTION_MODE_STATIC)
                        push_ui_command(CMD_PERFORM_ZERO);
                }
                else
                    send_button_event(EVENT_EXIT_MENU);
                break;
            default:
                break;
        }
    }
}

void UI::push_ui_command(UI::ui_cmd_t cmd) {
    _cmd_queue.try_add(cmd);
}

bool UI::get_pending_command(UI::ui_cmd_t &cmd) {
    return _cmd_queue.try_remove(cmd);
}

void UI::set_led_bar_level(int8_t level, color_map_t colormap) {
    uint8_t lev_ind = std::clamp((int)level, 0, 9);

    for (int i = 0; i < 10; i++) {
        uint8_t r=0, g=0, b=0;
        if (i < level) {
            if (colormap == COLOR_MAP_RED_TO_GREEN_SAME)
                fillColormapPixelValue(lev_ind, colormap, r, g, b);
            else
                fillColormapPixelValue(i, colormap, r, g, b);
        }
        set_pixel_data(i, r, g, b);
    }

    if (BOARD_HAS_BACKLIGHT && _backlight_enabled) {
        uint8_t v = _current_backlight_brightness_led_value();
        for (int i=0; i < 4; i++)
            set_pixel_data(10+i, 0, v, 0);
    } else {
        for (int i=0; i < 4; i++)
            set_pixel_data(10+i, 0, 0, 0);
    }

    transmit_pixel_data();
}

void UI::inc_setting(int8_t &set) {
    set++;
    if (set > 10)
        set = 10;
}

void UI::dec_setting(int8_t &set, bool zero_inclusive) {
    set--;
    if (zero_inclusive && set < 0) {
        set = 0;
    } else if (!zero_inclusive && set < 1) {
        set = 1;
    }
}

void UI::set_led_bar_amplitude(float amplitude, color_map_t colormap) {
    if (!in_menu()) {
        auto level = (uint8_t) ceilf(amplitude * NUM_WS2812_PIXELS);
        set_led_bar_level(level, colormap);
    }
}

void UI::updateVolume() {
    _audio.setVolume(get_volume_setting() / 10.0f);
}

void UI::updateVibroPower() {
    _vibro.set_max_power(get_vibro_setting() / 10.0f);
}

uint8_t UI::_current_backlight_brightness_led_value() {
    // mapping generated by octave: (brightness.* (sqrt(127)/10)).^2
    static const uint8_t exp_brightness_map[] = {0, 2, 5, 11, 20, 32, 45, 62, 81, 102, 127};
    if (_current_backlight_brightness <= 10)
        return exp_brightness_map[_current_backlight_brightness];
    else
        return 255;
}

void UI::startup_loop() {
#ifndef NDEBUG
    _startupDone = true;
    set_led_bar_level(0);
    _startupButtonInhibitTs = time_us_64() + 5 * 1000 * 1000;
    return;
#endif

    enum {
        INIT=0,
        WAIT_FIRST_RELEASE,
        WAIT_SECOND_RELEASE
    } startup_state = INIT;
    uint32_t pwr_cnt=0;
    uint32_t timeout_cnt=0;
    while (true) {

        switch (startup_state) {
            case INIT:
                startup_state = WAIT_FIRST_RELEASE;
                pwr_cnt = 0;
                break;
            case WAIT_FIRST_RELEASE:
                set_led_bar_level(10);
                if (get_button_state(BUTTON_POWER_BACK))
                    pwr_cnt++;
                else {
                    pwr_cnt = 0;
                    startup_state = WAIT_SECOND_RELEASE;
                }
                if (pwr_cnt > 1000) {
                    //enter bootloader mode
                    _audio.enableOutputDClass();
                    sleep_ms(100);
                    _audio.playSystemSound(PWMAudio::SOUND_FIRMWARE_UPDATE_MODE);
                    _audio.waitPlaybackFinished();
                    _audio.disableOutput();
                    reset_usb_boot(0, 0);
                }
                break;
            case WAIT_SECOND_RELEASE:
                auto level = pwr_cnt / 10;
                if (level > 10)
                    level = 10;
                set_led_bar_level(level);
                if (level == 10) {
                    _startupDone = true;
                    set_led_bar_level(0);
                    _startupButtonInhibitTs = time_us_64() + 5 * 1000 * 1000;
                    return;
                }
                if (get_button_state(BUTTON_POWER_BACK))
                    pwr_cnt++;
                else {
                    timeout_cnt++;
                    if (pwr_cnt != 0) {
                        if (pwr_cnt < 1000) {
                            //power off
                            gpio_put(PIN_POWER_EN, false);
                        } else {
                       }
                    }
                    if (timeout_cnt > 50) {
                        gpio_put(PIN_POWER_EN, false);
                    }
                }
                break;
        }
        sleep_ms(10);
    }
}

void UI::fillColormapPixelValue(uint8_t pixel, UI::color_map_t colormap, uint8_t &r, uint8_t &g, uint8_t &b) {
    if (colormap == COLOR_MAP_WHITE) {
        r = _current_backlight_brightness_led_value();
        g = r;
        b = r;
        return;
    }
    const uint8_t* px_ptr = nullptr;
    switch (colormap) {
        case COLOR_MAP_GREEN_TO_RED:
            px_ptr = LEVEL_MAP_GREEN_TO_RED[pixel];
            break;
        case COLOR_MAP_RED_TO_GREEN:
        case COLOR_MAP_RED_TO_GREEN_SAME:
            px_ptr = LEVEL_MAP_RED_TO_GREEN[pixel];
            break;
    }
    if (px_ptr) {
        auto brightness = _current_backlight_brightness_led_value();
        uint16_t r_ = px_ptr[0];
        uint16_t g_ = px_ptr[1];
        uint16_t b_ = px_ptr[2];
        r = r_ * brightness / 255;
        g = g_ * brightness / 255;
        b = b_ * brightness / 255;
    }
}


