//
// Created by R on 23.18.2.
//

#ifndef PIFW_UI_H
#define PIFW_UI_H

#include <pico/stdlib.h>
#include <pico/util/queue.h>
#include <hardware/pio.h>
#include <algorithm>
#include "audio/PWMAudio.h"
#include "util/Queue.h"
#include "structures.h"
#include "Vibro.h"
#include "HAL.h"

class UI {
public:
    UI(HAL &hal);
    bool init();

    /**
     * Loop that handles startup pattern / entering firmware update mode.
     * Returns when the rest of the unit can start up.
     */
    void startup_loop();

    typedef enum {
        BUTTON_POWER_BACK=0,
        BUTTON_UP,
        BUTTON_DOWN,
        BUTTON_OK_MENU,
        BUTTON_END
    } button_t;

    typedef enum {
        EVENT_POWER_ON=0,
        EVENT_POWER_OFF,
        EVENT_ENTER_MENU,
        EVENT_EXIT_MENU,
        EVENT_BACK,
        EVENT_UP,
        EVENT_DOWN,
        EVENT_END
    } event_t;
    typedef enum {
        POL_ACTIVE_HIGH = 0,
        POL_ACTIVE_LOW
    } button_polarity_t;

    typedef enum {
        CMD_PERFORM_ZERO=0,
        CMD_UPDATE_SETTINGS,
        CMD_INDICATE_VOLUME_CHANGE,
        CMD_INDICATE_VIBRO_POWER_CHANGE,
        CMD_INDICATE_POWER_OFF,
        CMD_ENABLE_PINPOINTING,
        CMD_DISABLE_PINPOINTING,
        CMD_DISABLE_DETECTION_PLAYBACK,
        CMD_INDICATE_BATTERY_LEVEL,
        CMD_ENTER_MENU,
        CMD_EXIT_MENU,
    } ui_cmd_t;

    typedef struct {
        uint8_t r;
        uint8_t g;
        uint8_t b;
    } pixel_t;

    void updateVolume();

    void updateVibroPower();
    bool get_button_state(button_t btn);

    /**
     * Sets pixel's RGB values in internal buffer (need to call transmit_pixel_data() afterwards)
     * @param pixel
     * @param r
     * @param g
     * @param b
     */
    void set_pixel_data(uint8_t pixel, uint8_t r, uint8_t g, uint8_t b);
    void transmit_pixel_data();

    void led_animation_blocking();
    void button_printf_state();

    typedef enum {
        COLOR_MAP_WHITE, // all levels white
        COLOR_MAP_GREEN_TO_RED, // green to red linear map
        COLOR_MAP_RED_TO_GREEN, //red to green linear map
        COLOR_MAP_RED_TO_GREEN_SAME,  //red to green, but color of every indicated LED is the same
    } color_map_t;

    void set_led_bar_level(int8_t indicated_level, color_map_t colormap = COLOR_MAP_WHITE);
    void set_led_bar_amplitude(float amplitude, color_map_t colormap);

    /**
     * Called from timer, handles button presses and generates events
     */
    void button_handler();

    /**
     * Send event to be processed by main loop
     * @param evt
     * @return true if event was accepted
     */
    bool    send_button_event(event_t evt);

    /**
     * Process events from user
     */
    void process_events();

    /**
     * Get a command from UI to process
     * @param cmd
     * @return
     */
    bool get_pending_command(ui_cmd_t& cmd);

    /**
     * Returns true if user is interacting with the menu/settings
     * @return
     */
    bool in_menu() const { return _selected_menu_item != MENU_NONE || _potentialShutdownCommand; }

    void set_volume_setting(int8_t s) { _current_volume = std::clamp(s, (int8_t)1, (int8_t)10);}
    void set_sensitivity_setting(int8_t s) { _current_sensitivity = std::clamp(s, (int8_t)1, (int8_t)10);}
    void set_brightness_setting(int8_t s) { _current_backlight_brightness = std::clamp(s, (int8_t)1, (int8_t)10);}

    void set_vibro_power_setting(int8_t s) { _current_vibro_power = std::clamp(s, (int8_t) 0, (int8_t) 10); }
    void set_backlight_setting(bool s) { _backlight_enabled = s; }
    void set_detection_mode_setting(detection_mode_t m) { _detectionMode = m; }

    int8_t get_volume_setting() const { return _current_volume; }
    int8_t get_sensitivity_setting() const { return _current_sensitivity; }
    int8_t get_brightness_setting() const { return _current_backlight_brightness; }

    int8_t get_vibro_setting() const { return _current_vibro_power; }
    bool get_backlight_setting() const { return _backlight_enabled; }
    detection_mode_t get_detection_mode_setting() const { return _detectionMode; }

    PWMAudio& getAudio() { return _audio; }

    void set_battery_soc(uint8_t soc) { _battery_soc = soc; }

    static constexpr uint8_t LEVEL_MAP_GREEN_TO_RED[10][3] = {
            {0, 255, 0},
            {64, 255, 0},
            {128, 255, 0},
            {192, 255, 0},
            {255, 255, 0},
            {255, 192, 0},
            {255, 128, 0},
            {255, 64, 0},
            {255, 32, 0},
            {255, 0, 0},
    };

    static constexpr uint8_t LEVEL_MAP_RED_TO_GREEN[10][3] = {
            {255, 0, 0},
            {255, 32, 0},
            {255, 64, 0},
            {255, 128, 0},
            {255, 192, 0},
            {255, 255, 0},
            {192, 255, 0},
            {128, 255, 0},
            {64, 255, 0},
            {0, 255, 0},
    };

protected:
    void _init_gpio();
    void _init_pio();
    void push_ui_command(ui_cmd_t cmd);
    static inline uint32_t urgb_u32(uint8_t r, uint8_t g, uint8_t b) {
        return
                ((uint32_t) (r) << 8) |
                ((uint32_t) (g) << 16) |
                (uint32_t) (b);
    }

    static const uint32_t PIN_WS2812=9;
    static const uint32_t NUM_WS2812_PIXELS=14;
    uint32_t              num_ws2812_active_pixels;
    static const uint32_t PIN_POWER_EN=7;

    uint32_t _pio_ws2812_sm;
    pio_hw_t* _pio_ws2812;
    static const uint32_t PIN_BUTTON[BUTTON_END];
    static const button_polarity_t PIN_BUTTON_POLARITY[BUTTON_END];
    repeating_timer_t _timer_buttons;
    static const uint32_t BUTTON_HANDLER_FREQ_HZ=100;

    bool        _button_prev_state[BUTTON_END];
    uint64_t    _button_prev_state_ts[BUTTON_END];
    /**
     *  This queue receives events generated from button_handler() for processing outside interrupt context
     *  (e.g. main loop).
     *
     *  Queue elements are of type event_t
     */
    Queue<event_t, 4> _button_event_queue;

    /**
     * This queue holds commands to be sent from UI to main loop
     */
    Queue<ui_cmd_t, 4> _cmd_queue;

    typedef enum {
        MENU_NONE=0,
        MENU_SENSITIVITY,
        MENU_VOLUME,
        MENU_VIBRO_POWER,
        MENU_BRIGHTNESS,
        MENU_BACKLIGHT,
        MENU_DETECTION_MODE
    } menu_t;

    menu_t _selected_menu_item;
    int8_t _current_sensitivity;
    int8_t _current_volume;
    int8_t _current_backlight_brightness;
    int8_t _current_vibro_power;
    bool   _backlight_enabled;
    uint8_t _current_backlight_brightness_led_value();
    void inc_setting(int8_t& set);

    void dec_setting(int8_t &set, bool zero_inclusive = false);
    bool _in_menu;
    const uint32_t MENU_TIMEOUT_US = 10 * 1000 * 1000;

    // how long does the user need to press power button to start turn-off LED animation
    const uint32_t POWER_OFF_INDICATION_START_MS = 1000;
    // how long does the user need to press power button to turn off
    const uint32_t POWER_OFF_PRESS_DURATION_MS = 3000;

    pixel_t _pixel_buffer[NUM_WS2812_PIXELS];
    uint32_t _pixel_dma_buffer[NUM_WS2812_PIXELS];
    int     _pixel_dma_chan;
    HAL&    _hal;
    PWMAudio& _audio;
    Vibro&   _vibro;
    bool    _poweringOff;
    detection_mode_t _detectionMode;

    bool    _startupDone;
    uint64_t _startupButtonInhibitTs;

    uint8_t _battery_soc;
    bool   _potentialShutdownCommand;
    void fillColormapPixelValue(uint8_t pixel, color_map_t colormap, uint8_t& r, uint8_t& g, uint8_t& b);
};


#endif //PIFW_UI_H
