//
// Created by R on 23.19.2.
//

#include "PWMAudio.h"
#include <pico/stdlib.h>
#include <hardware/pwm.h>
#include <cmath>
#include "audio.h"
#include "../util/array.h"

//auto-generated headers from wav files
#include <audio/samples.h>
#include <cstring>
#include "../util/Note.h"
#include "../util/array.h"

#define _SOUND_KEY(k) SOUND_##k
#define SOUND_KEY(k) _SOUND_KEY(k)
#define _WAV_KEY(k)  WAV_DATA_##k
#define WAV_KEY(k)   _WAV_KEY(k)

#define SOUND_ITEM(k) {SOUND_KEY(k), WAV_KEY(k), NUM_ELEMENTS(WAV_KEY(k))}

const PWMAudio::system_sound_t PWMAudio::SYSTEM_SOUNDS[] = {
        SOUND_ITEM(INITIALIZING),
        SOUND_ITEM(SENSITIVITY_ADJUSTMENT),
        SOUND_ITEM(VOLUME_ADJUSTMENT),
        SOUND_ITEM(AUTO_OFFSET_FAILED),
        SOUND_ITEM(AUTO_OFFSET_FAILED),
        SOUND_ITEM(REF_SAMPLE_UPDATED),
        SOUND_ITEM(BRIGHTNESS_ADJUSTMENT),
        SOUND_ITEM(BACKLIGHT),
        SOUND_ITEM(ON),
        SOUND_ITEM(OFF),
        SOUND_ITEM(STATIC),
        SOUND_ITEM(DYNAMIC),
        SOUND_ITEM(DETECTION_MODE),
        SOUND_ITEM(FIRMWARE_UPDATE_MODE),
        SOUND_ITEM(BATTERY_LEVEL_HIGH),
        SOUND_ITEM(BATTERY_LEVEL_MEDIUM),
        SOUND_ITEM(BATTERY_LEVEL_LOW),
        SOUND_ITEM(BATTERY_LEVEL_CRITICAL),
        SOUND_ITEM(VIBRATION_POWER_ADJUSTMENT),
        SOUND_ITEM(RESTORE_DEFAULT_SETTINGS),
        SOUND_ITEM(DEFAULT_SETTINGS_RESTORED),
        SOUND_ITEM(SIGNAL_CLIPPING_DETECTED)
};

static const note_t INIT_NOTES[] = {
        {NOTE_B4, 120},
        {0, 75}

};
const melody_t MELODY_INIT = {
        .notes = reinterpret_cast<const note_t *>(&INIT_NOTES),
        .num_notes = NUM_ELEMENTS(INIT_NOTES),
        .repeat = true
};

static const note_t CONFIRM_NOTES[] = {
        {NOTE_B4, 100},
        {0, 100},
        {NOTE_B5, 100}
};

const melody_t MELODY_CONFIRM = {
        .notes = reinterpret_cast<const note_t *>(&CONFIRM_NOTES),
        .num_notes = NUM_ELEMENTS(CONFIRM_NOTES),
        .repeat = false
};
static const note_t SHORT_BEEP_NOTES[] = {
        {NOTE_B4, 100}
};

const melody_t MELODY_SHORT_BEEP = {
        .notes = reinterpret_cast<const note_t *>(&SHORT_BEEP_NOTES),
        .num_notes = NUM_ELEMENTS(SHORT_BEEP_NOTES),
        .repeat = false
};

static const note_t POWER_OFF_NOTES[] = {
        {NOTE_B4, 250},
        {0, 20},
        {NOTE_B3, 300},
        {0, 20},
        {NOTE_B2, 350},
        {0, 20},
};

const melody_t MELODY_POWER_OFF= {
        .notes = reinterpret_cast<const note_t *>(&POWER_OFF_NOTES),
        .num_notes = NUM_ELEMENTS(POWER_OFF_NOTES),
        .repeat = false
};


static PWMAudio* INST = nullptr;

void headphone_detect_gpio_callback(uint gpio, uint32_t events) {
    if (INST && gpio == PWMAudio::HEADPHONE_DETECT_PIN) {
        INST->clearHeadphonesDetectedRuntimeFlag();
        gpio_set_irq_enabled_with_callback(PWMAudio::HEADPHONE_DETECT_PIN,
                                           GPIO_IRQ_EDGE_FALL, false, &headphone_detect_gpio_callback);
    }
}

PWMAudio::PWMAudio() :
        _volume(0.1f),
        _nextSineBuffer(0),
        _headphonesDetected(false),
        _headphonesDetectedRuntime(false),
        _headphonesDetectedRuntimeFlag(false),
        _playbackState(PLAYBACK_STOPPED),
        _currentOutputFreq(0),
        _targetOutputFreq(0),
        _noOutputCounter(0),
        _playbackStepCounter(0)
{

    memset(&melody_playback, 0, sizeof(melody_playback));
}

void PWMAudio::init() {
    gpio_init(EN_PIN);
    gpio_set_dir(EN_PIN, true);
    gpio_put(EN_PIN, false);
    gpio_init(HEADPHONE_DETECT_PIN);
    gpio_set_dir(HEADPHONE_DETECT_PIN, false);
    detectHeadphones();
    audio_init(PWM_PIN, 11000);
    uint32_t longest_system_sound = 0;
    for (auto&s:SYSTEM_SOUNDS) {
        if (s.length > longest_system_sound)
            longest_system_sound = s.length;
    }
    _volumeAdjustmentBuffer = (uint8_t*)malloc(longest_system_sound);
    _initialized = true;
}


void PWMAudio::playBuffer(uint8_t *buffer, uint32_t size, bool repeat) {
    audio_play(buffer, size, repeat);
    _currentSineFreq = 0;
}

float fastsine(float x)
{
    const float B = 4/M_PI;
    const float C = -4/(M_PI*M_PI);

    return B * x + C * x * fabs(x);
}

void PWMAudio::playSineWave(int16_t freq_hz, bool repeat) {
    //first fill _sineBuffer with desired sinewave
    if (freq_hz < 200 || freq_hz > 5500)
        return;

    uint32_t sine_len_samples = 1.0f / (float)freq_hz * 11000.0f;
    assert(sine_len_samples < SINE_LEN);
    float sine_step = 2.0f * M_PI / sine_len_samples;

    float r = 0.0f;

//    unsigned int num_repeat = sizeof(_sineBuffer) / sine_len_samples;
    uint8_t* sinbuf = &(_sineBuffer[_nextSineBuffer][0]);
    _nextSineBuffer++;
    if (_nextSineBuffer > 1)
        _nextSineBuffer = 0;
    for (int i=0; i < sine_len_samples; i++) {
        sinbuf[i] = 127 + sinf(r) * 127.0f * _volume;
        r += sine_step;
    }

    playBuffer(sinbuf, sine_len_samples, repeat);
    _currentSineFreq = freq_hz;
}

void PWMAudio::enableOutputABClass() {
    if (!headphonesDetected())
        gpio_put(EN_PIN, true);

    _outputEnabled = true;
}

void PWMAudio::enableOutputDClass() {
    if (!headphonesDetected()) {
        gpio_put(EN_PIN, false);
        busy_wait_us(100);
        gpio_put(EN_PIN, true);
        busy_wait_us(3);
        gpio_put(EN_PIN, false);
        busy_wait_us(3);
        gpio_put(EN_PIN, true);
        busy_wait_ms(10);
    }
    _outputEnabled = true;
}

void PWMAudio::disableOutput() {
    gpio_put(EN_PIN, false);
    _outputEnabled = false;
}

void PWMAudio::stopPlayback() {
    audio_stop();
}

void PWMAudio::playSystemSound(PWMAudio::system_sound_key_t sound) {
    hard_assert(_volumeAdjustmentBuffer);

    int32_t vol_mul = _volume * 256;
    for (auto&ss: SYSTEM_SOUNDS) {
        if (ss.key == sound) {

            for (int i=0; i < ss.length; i++) {
                int32_t a = ss.ptr[i];

                a = (((a - 128) * vol_mul) >> 8) + 128;
                if (a>255)
                    a=255;
                _volumeAdjustmentBuffer[i] = a;
            }
            playBuffer(_volumeAdjustmentBuffer, ss.length, false);
            _playbackState = PLAYBACK_SYSTEM_SOUND;
            _playbackStepCounter = 0;
        }
    }
}

void PWMAudio::waitPlaybackFinished() {
    audio_wait_playback_finished();
}

bool PWMAudio::isPlaybackFinished() {
    return !audio_is_playing();
}

bool PWMAudio::headphonesDetected() {
    return _headphonesDetected;
}

void PWMAudio::detectHeadphones() {
    gpio_init(PWM_PIN);
    gpio_set_dir(PWM_PIN, true);
    gpio_put(PWM_PIN, true);
    sleep_ms(5);
    _headphonesDetected = gpio_get(HEADPHONE_DETECT_PIN);
}

void PWMAudio::enableRuntimeHeadphoneDetection() {
    INST = this;
    _headphonesDetectedRuntimeFlag = true;
    gpio_set_irq_enabled_with_callback(HEADPHONE_DETECT_PIN,
                                       GPIO_IRQ_EDGE_FALL, true, &headphone_detect_gpio_callback);
}

void PWMAudio::disableRuntimeHeadphoneDetection() {
    auto prevHeadphonesDetected = _headphonesDetected;

    _headphonesDetected = _headphonesDetectedRuntimeFlag;

    gpio_set_irq_enabled_with_callback(HEADPHONE_DETECT_PIN,
                                       GPIO_IRQ_EDGE_FALL, false, &headphone_detect_gpio_callback);

    if (!prevHeadphonesDetected && _headphonesDetected)
        enableOutputDClass();
    else if (prevHeadphonesDetected && !_headphonesDetected)
        disableOutput();
}

void PWMAudio::step() {
    switch (_playbackState) {
        case PLAYBACK_STOPPED:
            break;
        case PLAYBACK_SYSTEM_SOUND:
            _playbackStepCounter++;
            if (_playbackStepCounter > 30 && !audio_is_playing())
                _playbackState = PLAYBACK_STOPPED;

            break;
        case PLAYBACK_DETECTION:
            stepPlaybackDetection();
            break;
        case PLAYBACK_MELODY:
            stepPlaybackMelody();
            break;
    }
}

void PWMAudio::playInitTune() {
    playMelody(MELODY_INIT);
}

void PWMAudio::stepPlaybackMelody() {
    if (melody_playback.melody.notes == nullptr)
        return;

    if (!isOutputEnabled()) {
        enableOutputDClass();
    }

    if (melody_playback.current_note == -1 || (melody_playback.current_note_end_us < time_us_64())) {
        melody_playback.current_note++;
        auto note = melody_playback.melody.notes[melody_playback.current_note];
        melody_playback.current_note_end_us = time_us_64() + note.duration_ms * 1000;
    }

    if (melody_playback.current_note >= melody_playback.melody.num_notes) {
        //final note finished, check if need to repeat
        if (melody_playback.melody.repeat) {
            melody_playback.current_note = 0;
            auto note = melody_playback.melody.notes[melody_playback.current_note];
            melody_playback.current_note_end_us = time_us_64() + note.duration_ms * 1000;
        } else {
            stopPlayback();
            melody_playback.melody.notes = nullptr;
            _playbackState = PLAYBACK_STOPPED;
        }
    } else {
        auto note = melody_playback.melody.notes[melody_playback.current_note];
        if (note.freq != 0) {
            playSineWave(note.freq);
        } else {
            stopPlayback();
        }
    }
}

void PWMAudio::playMelody(melody_t melody) {
    melody_playback.melody = melody;
    melody_playback.current_note = -1;
    _playbackState = PLAYBACK_MELODY;
}

bool PWMAudio::isMelodyPlaying() const {
    return _playbackState == PLAYBACK_MELODY && melody_playback.melody.notes != nullptr;
}

void PWMAudio::waitMelodyPlaybackFinished() {
    while (isMelodyPlaying()) {
        sleep_ms(1);
    }
}

void PWMAudio::stepPlaybackDetection() {
    const int16_t FREQ_STEP = 10;

    if (_currentOutputFreq < _targetOutputFreq) {
        _currentOutputFreq += FREQ_STEP;
        if (_currentOutputFreq > _targetOutputFreq)
            _currentOutputFreq = _targetOutputFreq;
    } else if (_currentOutputFreq > _targetOutputFreq) {
        _currentOutputFreq -= FREQ_STEP;
        if (_currentOutputFreq < _targetOutputFreq)
            _currentOutputFreq = _targetOutputFreq;
    }

    if (_currentOutputFreq > 0) {
        if (!isOutputEnabled()) {
            enableOutputDClass();
        }

        if (getCurrentSineWaveFrequency() != _currentOutputFreq) {
            playSineWave(_currentOutputFreq, true);
        }
        _noOutputCounter = 0;
    } else {
        uint64_t current_time_us = time_us_64();
        if (_noOutputCounter == 0 || current_time_us > (_nextConfidenceClickAtUs + CONFIDENCE_CLICK_DURATION_US)) {
            stopPlayback();
            playSineWave(0, false);
            disableRuntimeHeadphoneDetection();
            _nextConfidenceClickAtUs = current_time_us + CONFIDENCE_CLICK_DELAY_US;
        } else if ((_nextConfidenceClickAtUs < current_time_us) && (current_time_us < (_nextConfidenceClickAtUs + CONFIDENCE_CLICK_DURATION_US))) {
            if (getCurrentSineWaveFrequency() != _currentOutputFreq) {
                playSineWave(CONFIDENCE_CLICK_TONE_HZ, true);
                enableRuntimeHeadphoneDetection();
            }
        }

        _noOutputCounter++;
    }
}

bool PWMAudio::isSystemSoundPlaying() const {
    return _playbackState == PLAYBACK_SYSTEM_SOUND;
}

void PWMAudio::waitSystemSoundPlaybackFinished() {
    while (isSystemSoundPlaying())
        sleep_ms(1);
}

void PWMAudio::playDetectionTone(int16_t freq) {
    _playbackState = PLAYBACK_DETECTION;
    _targetOutputFreq = freq;
}
