//
// Created by R on 23.19.2.
//

#ifndef PIFW_PWMAUDIO_H
#define PIFW_PWMAUDIO_H

#include <cstdint>
#include <cstddef>
#include <pico/time.h>
#include <cstdio>

typedef struct {
    uint16_t freq;
    uint32_t duration_ms;
} note_t;

typedef struct {
    const note_t*     notes;
    size_t      num_notes;
    bool        repeat;
} melody_t;

extern const melody_t MELODY_INIT;
extern const melody_t MELODY_CONFIRM;
extern const melody_t MELODY_SHORT_BEEP;
extern const melody_t MELODY_POWER_OFF;


/**
 * This class is responsible for several things:
 * Managing audio output (enable/disable audio amplifier)
 * Detect headphone connection (also enables/disables audio amplifier)
 * Playing back sounds (sine wave, system sounds stored in flash).
 *
 * Headphone detection logic:
 * The headphone detection is done differently on startup and when sound is being played.
 * On startup, the audio output pin is set to high level and the input is monitored (it's low when headphones are connected)
 * When audio is already playing, the presence of headphones can be detected by observing PWM signal on the detect pin
 * when headphones are not connected, and a solid high signal when headphones are connected.
 *
 */
class PWMAudio {
public:
    PWMAudio();

    void playBuffer(uint8_t *buffer, uint32_t size, bool repeat=false);
    /**
     * Plays sine wave of the given frequency
     * @param freq_hz
     */
    void playSineWave(int16_t freq_hz, bool repeat=true);
    void init();
    bool isInitialized() const { return _initialized; }

    /**
     * Sets volume 0..1.0f
     * @param volume
     */
    void setVolume(const float& volume) { _volume = volume; }
    static void stopPlayback();
    void enableOutputABClass();
    void enableOutputDClass();
    void disableOutput();

    bool isOutputEnabled() const { return _outputEnabled; }

    int16_t getCurrentSineWaveFrequency() const { return _currentSineFreq; }

    bool headphonesDetected();

    void enableRuntimeHeadphoneDetection();
    void disableRuntimeHeadphoneDetection();

    void playDetectionTone(int16_t freq);

    typedef enum {
        SOUND_INITIALIZING=0,
        SOUND_VOLUME_ADJUSTMENT,
        SOUND_SENSITIVITY_ADJUSTMENT,
        SOUND_AUTO_OFFSET_FAILED,
        SOUND_REF_SAMPLE_UPDATED,
        SOUND_BRIGHTNESS_ADJUSTMENT,
        SOUND_BACKLIGHT,
        SOUND_ON,
        SOUND_OFF,
        SOUND_DYNAMIC,
        SOUND_STATIC,
        SOUND_DETECTION_MODE,
        SOUND_FIRMWARE_UPDATE_MODE,
        SOUND_BATTERY_LEVEL_HIGH,
        SOUND_BATTERY_LEVEL_MEDIUM,
        SOUND_BATTERY_LEVEL_LOW,
        SOUND_BATTERY_LEVEL_CRITICAL,
        SOUND_VIBRATION_POWER_ADJUSTMENT,
        SOUND_RESTORE_DEFAULT_SETTINGS,
        SOUND_DEFAULT_SETTINGS_RESTORED,
        SOUND_SIGNAL_CLIPPING_DETECTED
    } system_sound_key_t;

    typedef struct {
        system_sound_key_t key;
        const uint8_t*           ptr;
        uint32_t           length;
    } system_sound_t;

    void playSystemSound(system_sound_key_t sound);
    void playMelody(melody_t melody);

    bool isSystemSoundPlaying() const;
    void waitSystemSoundPlaybackFinished();

    bool isMelodyPlaying() const;
    void waitMelodyPlaybackFinished();

    void waitPlaybackFinished();
    bool isPlaybackFinished();

    static const uint32_t HEADPHONE_DETECT_PIN = 20;

    void clearHeadphonesDetectedRuntimeFlag() { _headphonesDetectedRuntimeFlag = false; }

    void step();


    typedef enum {
        PLAYBACK_STOPPED=0,
        PLAYBACK_MELODY,
        PLAYBACK_SYSTEM_SOUND,
        PLAYBACK_DETECTION,
    } playback_state_t;

    playback_state_t getPlaybackState() const { return _playbackState; }

    void setPlaybackState(playback_state_t s) { _playbackState = s; }

    void playInitTune();
    void stopPlayInitTune();



protected:
    const uint64_t CONFIDENCE_CLICK_DURATION_US = (10 * 1000);
    const uint64_t CONFIDENCE_CLICK_DELAY_US = (1 * 1000 * 1000);
    const uint16_t CONFIDENCE_CLICK_TONE_HZ = 440;
    int16_t     _currentOutputFreq;
    int16_t     _targetOutputFreq;
    uint32_t    _noOutputCounter;
    uint64_t    _nextConfidenceClickAtUs;
    void stepPlaybackDetection();
    void stepPlaybackMelody();
    void stepPlaybackSystemSound();
    void stepPlaybackConfidenceClick();

    static const uint32_t PWM_PIN=19;
    static const uint32_t EN_PIN=18;
    static const uint32_t  SINE_LEN=64; //1/10th of sample rate
    uint8_t     _sineBuffer[2][SINE_LEN];
    uint8_t*    _volumeAdjustmentBuffer;
    uint32_t    _nextSineBuffer;
    float _volume;
    int16_t _currentSineFreq;
    bool _outputEnabled;

    static const system_sound_t SYSTEM_SOUNDS[];

    bool _headphonesDetected;
    bool _headphonesDetectedRuntimeFlag;
    bool _headphonesDetectedRuntime;
    void detectHeadphones();

    struct {
        melody_t melody;
        ssize_t current_note;
        uint64_t current_note_end_us;
    } melody_playback;

    playback_state_t _playbackState;

    bool _initialized;

    uint32_t _playbackStepCounter;
};


#endif //PIFW_PWMAUDIO_H
