#include <cstring>
#include <cstdio>

#include <pico/stdlib.h>
#include <hardware/dma.h>
#include <hardware/irq.h>
#include <hardware/pwm.h>
#include <hardware/sync.h>
#include <hardware/clocks.h>
#include <pico/util/queue.h>

#include "audio.h"
#include <pico/sync.h>

#define REPETITION_RATE 4

/**
 * Code organization: due to strict timing requirements of the audio playback, the audio playback is handled by 2nd core.
 * audio_init() should be executed in the second core, after which audio_playback_step(); should be called in a loop.
 *
 * audio_play() and stop can be then executed on the first core, which will use queue to communicate with audio_playback_step()
 *
 * @TODO: Implement double buffering
 */

static int pwm_dma_chan, trigger_dma_chan, sample_dma_chan;
static critical_section_t dma_critical_section;
static uint32_t single_sample = 0;
static uint32_t *single_sample_ptr = &single_sample;
static volatile uint32_t next_dma_read_buf=0;
typedef struct {
    const uint8_t* buffer;
    uint32_t length;
    bool repeat;
    bool playback_done;
} audio_buffer_t;
static volatile audio_buffer_t audio_buffer[3];

typedef struct {
    uint32_t buffer_length;
    const uint8_t* buffer;
    bool repeat;
} playback_item_t;

static queue_t playbackQueue;
static void __no_inline_not_in_flash_func(dma_handler)()
{
    if (audio_buffer[next_dma_read_buf].playback_done && !audio_buffer[next_dma_read_buf].repeat) {

        next_dma_read_buf = 0;

    }
    dma_hw->ch[sample_dma_chan].al1_read_addr       = (intptr_t) audio_buffer[next_dma_read_buf].buffer;
    dma_hw->ch[trigger_dma_chan].transfer_count = audio_buffer[next_dma_read_buf].length;
    dma_hw->ch[trigger_dma_chan].al3_read_addr_trig = (intptr_t) &single_sample_ptr;
    audio_buffer[next_dma_read_buf].playback_done = true;

    dma_hw->ints1 = 1u << trigger_dma_chan;
}

static uint8_t silence[64];
void audio_init(int audio_pin, int sample_freq)
{
    queue_init(&playbackQueue, sizeof(playback_item_t), 16);
    memset((void*)audio_buffer, 0, sizeof(audio_buffer));
    memset(silence, 127, sizeof(silence));
    audio_buffer[0].buffer = silence;
    audio_buffer[0].length = REPETITION_RATE * sizeof(silence);
    audio_buffer[0].repeat = true;
    audio_buffer[0].playback_done = false;
    critical_section_init(&dma_critical_section);
    gpio_set_function(audio_pin, GPIO_FUNC_PWM);

    int audio_pin_slice = pwm_gpio_to_slice_num(audio_pin);
    int audio_pin_chan = pwm_gpio_to_channel(audio_pin);

    uint f_clk_sys = frequency_count_khz(CLOCKS_FC0_SRC_VALUE_CLK_SYS);
    float clock_div = ((float)f_clk_sys * 1000.0f) / 254.0f / (float) sample_freq / (float) REPETITION_RATE;

    pwm_config config = pwm_get_default_config();
    pwm_config_set_clkdiv(&config, clock_div);
    pwm_config_set_wrap(&config, 254);
    pwm_init(audio_pin_slice, &config, true);

    pwm_dma_chan     = dma_claim_unused_channel(true);
    trigger_dma_chan = dma_claim_unused_channel(true);
    sample_dma_chan  = dma_claim_unused_channel(true);

    // setup PWM DMA channel
    dma_channel_config pwm_dma_chan_config = dma_channel_get_default_config(pwm_dma_chan);
    channel_config_set_transfer_data_size(&pwm_dma_chan_config, DMA_SIZE_32);              // transfer 32 bits at a time
    channel_config_set_read_increment(&pwm_dma_chan_config, false);                        // always read from the same address
    channel_config_set_write_increment(&pwm_dma_chan_config, false);                       // always write to the same address
    channel_config_set_chain_to(&pwm_dma_chan_config, sample_dma_chan);                    // trigger sample DMA channel when done
    channel_config_set_dreq(&pwm_dma_chan_config, DREQ_PWM_WRAP0 + audio_pin_slice);       // transfer on PWM cycle end
    dma_channel_configure(pwm_dma_chan,
                          &pwm_dma_chan_config,
                          &pwm_hw->slice[audio_pin_slice].cc,   // write to PWM slice CC register
                          &single_sample,                       // read from single_sample
                          REPETITION_RATE,                      // transfer once per desired sample repetition
                          false                                 // don't start yet
    );

    // setup trigger DMA channel
    dma_channel_config trigger_dma_chan_config = dma_channel_get_default_config(trigger_dma_chan);
    channel_config_set_transfer_data_size(&trigger_dma_chan_config, DMA_SIZE_32);          // transfer 32-bits at a time
    channel_config_set_read_increment(&trigger_dma_chan_config, false);                    // always read from the same address
    channel_config_set_write_increment(&trigger_dma_chan_config, false);                   // always write to the same address
    channel_config_set_dreq(&trigger_dma_chan_config, DREQ_PWM_WRAP0 + audio_pin_slice);   // transfer on PWM cycle end
    dma_channel_configure(trigger_dma_chan,
                          &trigger_dma_chan_config,
                          &dma_hw->ch[pwm_dma_chan].al3_read_addr_trig,     // write to PWM DMA channel read address trigger
                          &single_sample_ptr,                               // read from location containing the address of single_sample
                          audio_buffer[next_dma_read_buf].length,              // trigger once per audio sample per repetition rate
                          false                                             // don't start yet
    );
    dma_channel_set_irq1_enabled(trigger_dma_chan, true);    // fire interrupt when trigger DMA channel is done
    irq_set_exclusive_handler(DMA_IRQ_1, dma_handler);
    irq_set_enabled(DMA_IRQ_1, true);
    irq_set_priority(DMA_IRQ_1, 1);
    // setup sample DMA channel
    dma_channel_config sample_dma_chan_config = dma_channel_get_default_config(sample_dma_chan);
    channel_config_set_transfer_data_size(&sample_dma_chan_config, DMA_SIZE_8);  // transfer 8-bits at a time
    channel_config_set_read_increment(&sample_dma_chan_config, true);            // increment read address to go through audio buffer
    channel_config_set_write_increment(&sample_dma_chan_config, false);          // always write to the same address
    dma_channel_configure(sample_dma_chan,
                          &sample_dma_chan_config,
                          (char*)&single_sample + 2*audio_pin_chan,  // write to single_sample
                          audio_buffer[next_dma_read_buf].buffer,                      // read from audio buffer
                          1,                                         // only do one transfer (once per PWM DMA completion due to chaining)
                          false                                      // don't start yet
    );


    // kick things off with the trigger DMA channel
    dma_channel_start(trigger_dma_chan);
}

bool audio_play(const uint8_t *samples, int len, bool repeat) {

    playback_item_t pi;
    pi.repeat = repeat;
    pi.buffer_length = len;
    pi.buffer = samples;

    if (!queue_try_add(&playbackQueue, &pi)) {
        busy_wait_at_least_cycles(1);
        return false;
    }
    return true;
}

void audio_stop() {

    playback_item_t pi;

    //while (queue_try_remove(&playbackQueue, &pi))
    //    tight_loop_contents();

    pi.repeat = false;
    pi.buffer = nullptr;
    pi.buffer_length = 0;
    queue_try_add(&playbackQueue, &pi);
}

void __no_inline_not_in_flash_func(audio_playback_step)() {
    playback_item_t pi;
    if (queue_try_remove(&playbackQueue, &pi)) {
        if (pi.buffer && pi.buffer_length) {
            auto status = save_and_disable_interrupts();
            uint32_t nidx;
            if  (next_dma_read_buf != 1)
                nidx = 1;
            else
                nidx = 2;

            audio_buffer[nidx].length = pi.buffer_length * REPETITION_RATE;
            audio_buffer[nidx].buffer = pi.buffer;
            audio_buffer[nidx].repeat = pi.repeat;
            audio_buffer[nidx].playback_done = false;

            next_dma_read_buf = nidx;
            restore_interrupts(status);
        } else {
            //this is a stop command
            next_dma_read_buf = 0;

            dma_channel_set_irq1_enabled(trigger_dma_chan, false);
            dma_channel_abort(trigger_dma_chan);
            dma_channel_acknowledge_irq1(trigger_dma_chan);
            dma_hw->ch[sample_dma_chan].al1_read_addr       = (intptr_t) audio_buffer[next_dma_read_buf].buffer;
            dma_hw->ch[trigger_dma_chan].transfer_count = audio_buffer[next_dma_read_buf].length;
            dma_hw->ch[trigger_dma_chan].al3_read_addr_trig = (intptr_t) &single_sample_ptr;
            audio_buffer[next_dma_read_buf].playback_done = true;

            dma_channel_set_irq1_enabled(trigger_dma_chan, true);
            dma_channel_start(trigger_dma_chan);
        }
    }
}

void audio_wait_playback_finished() {
    while (audio_is_playing())
        sleep_ms(1);
}

bool audio_is_playing() {
    return next_dma_read_buf != 0;
}




