//
// Created by R on 4/17/23.
//

#include "ADC.h"
#include <hardware/dma.h>
#include <hardware/clocks.h>
#include "spi_rx.pio.h"
#include "spi_reader_sm3.pio.h"

static ADC* INST=nullptr;

void dma_irq_handler() {
    INST->execute_dma_finished_callback();
}

ADC::ADC() : _dmaFinishedCallback(nullptr) {

    INST = this;
}

void ADC::init(PIO pio, uint32_t spi_sm) {

    _spi_pio = pio;
    _spi_sm = spi_sm;
    gpio_set_drive_strength(_gpio_cs, GPIO_DRIVE_STRENGTH_2MA);
    gpio_set_drive_strength(_gpio_sclk, GPIO_DRIVE_STRENGTH_2MA);

    float clkdiv = clock_get_hz(clk_sys) / _spi_freq_hz / 4.0f;
    auto spi_rx_prog_offs = pio_add_program(_spi_pio, &spi_rx_program);

    //perform input offset calibration on power up according to ADS7052
    //initialize the free-running SPI reader, read a single word
    pio_spi_rx_init(_spi_pio, _spi_sm, spi_rx_prog_offs, 24, clkdiv,
                 _gpio_sclk, _gpio_sdi, _gpio_cs);
    acquire_sample_blocking();
    pio_sm_set_enabled(_spi_pio, _spi_sm, false);
    pio_remove_program(_spi_pio, &spi_rx_program, spi_rx_prog_offs);
    //now initialize the real state machine
    auto spi_reader_prog_offs = pio_add_program(_spi_pio, &spi_reader_sm3_program);
    pio_spi_reader_sm3_init(_spi_pio, _spi_sm, spi_reader_prog_offs, 18, clkdiv,
                            _gpio_sclk, _gpio_sdi, _gpio_cs);

    _init_dma();
}

uint32_t ADC::acquire_sample_blocking() {
    uint32_t read_word=0xffff;
    read_word = pio_sm_get_blocking(_spi_pio, _spi_sm) >> 4;
    return read_word;
}

void ADC::_init_dma() {
    _dma_chan = dma_claim_unused_channel(true);
    auto c = dma_channel_get_default_config(_dma_chan);
    channel_config_set_transfer_data_size(&c, DMA_SIZE_32);
    channel_config_set_read_increment(&c, false);
    channel_config_set_write_increment(&c, true);
    channel_config_set_dreq(&c, pio_get_dreq(_spi_pio, _spi_sm, false));
    dma_channel_configure(_dma_chan,
                          &c, nullptr, &(_spi_pio->rxf[_spi_sm]), 0, false);
    dma_channel_set_irq0_enabled(_dma_chan, true);
    irq_set_exclusive_handler(DMA_IRQ_0, dma_irq_handler);
    irq_set_enabled(DMA_IRQ_0, true);
}

void ADC::acquire_samples_dma_blocking(uint32_t *buf, size_t num_samples) {
    acquire_samples_dma_start(buf, num_samples, false);
    dma_channel_wait_for_finish_blocking(_dma_chan);
    shift_dma_samples(buf, num_samples);
}


bool ADC::dma_transaction_finished() {
    return !dma_channel_is_busy(_dma_chan);
}

void ADC::shift_dma_samples(uint32_t *buf, size_t num_samples) {
    for (int i=0; i < num_samples; i++) {
        buf[i] = buf[i] >> 4;
    }
}

void ADC::register_dma_finished_callback(ADC::adc_callback_t fn) {
    _dmaFinishedCallback = fn;
}

void ADC::execute_dma_finished_callback() {
    dma_channel_acknowledge_irq0(_dma_chan);
    if (_dmaBuf != nullptr)
        shift_dma_samples(_dmaBuf, _dmaLen);
    if (_dmaFinishedCallback)
        _dmaFinishedCallback();
}
