from PyQt6 import QtWidgets, QtCore
import pi_pb2 as pipb
import logging
import typing

class SamplingConfigurationWidget(QtWidgets.QGroupBox):
    fieldWidgets: typing.Dict[str, QtWidgets.QLineEdit]

    configurationProvided = QtCore.pyqtSignal(object)
    configurationRequest = QtCore.pyqtSignal()
    acquireReferenceSampleRequest = QtCore.pyqtSignal()
    automaticOffsetAdjustmentRequest = QtCore.pyqtSignal()

    intFields = ["pulse_len_us", "pulse_period_us", "num_samples_analyze", "num_samples_skip",
                  "num_oversample_reference", "num_oversample_measurement", "num_undersample_processing",
                 "auto_offset_adjustment_range_min", "auto_offset_adjustment_range_max",
                 "ref_sample_update_period_us"]

    GAIN_LEVELS = pipb.SamplingConfiguration.FrontendGainLevel.items()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.log = logging.getLogger("samplingcfg")
        self.setTitle("Sampling configuration")
        self.fieldWidgets = {}

        layout = QtWidgets.QFormLayout()

        def cw(name, title, def_val):
            w = QtWidgets.QLineEdit(def_val)
            layout.addRow(title, w)
            self.fieldWidgets[name] = w

        cw("pulse_len_us", "Pulse length (us)", "100")
        cw("pulse_period_us", "Pulse period (us)", "10000")
        cw("frontend_manual_offset", "Frontend offset (0..1)", "0.666")

        w = QtWidgets.QCheckBox()
        w.setChecked(True)
        layout.addRow("Manual offset control", w)
        self.fieldWidgets["frontend_manual_offset_control"] = w

        cw("num_samples_analyze", "Num. samples to analyze", "16")
        cw("num_samples_skip", "Num. samples to skip", "7")
        cw("num_oversample_reference", "Oversample reference", "8")
        cw("num_oversample_measurement", "Oversample measurements", "4")
        cw("num_undersample_processing", "Process every Nth sample", "1")
        cw("ref_sample_update_period_us", "Ref. sample update period (us)", "800000")
        cw("auto_offset_adjustment_range_min", "Auto offset adj. target min", "12000")
        cw("auto_offset_adjustment_range_max", "Auto offset adj. target max", "15000")

        gcb = QtWidgets.QComboBox()
        for name, value in self.GAIN_LEVELS:
            gcb.addItem(name, value)
        self.fieldWidgets["gain_level"] = gcb
        layout.addRow("Frontend gain", gcb)

        wc = QtWidgets.QPushButton("Write configuration")
        wc.released.connect(self.handleWriteConfiguration)
        rc = QtWidgets.QPushButton("Read configuration")
        rc.released.connect(lambda: self.configurationRequest.emit())
        layout.addRow(rc, wc)

        rfr = QtWidgets.QPushButton("Acquire ref. sample (zero)")
        rfr.released.connect(lambda: self.acquireReferenceSampleRequest.emit())

        aoa = QtWidgets.QPushButton("Do auto offset adj")
        aoa.released.connect(lambda: self.automaticOffsetAdjustmentRequest.emit())

        layout.addRow(aoa, rfr)
        self.setLayout(layout)

    @QtCore.pyqtSlot()
    def handleWriteConfiguration(self):
        self.log.info("writing configuration")

        msg = pipb.SamplingConfiguration()
        for f in self.intFields:
            setattr(msg, f, int(self.fieldWidgets[f].text()))
        msg.frontend_manual_offset = float(self.fieldWidgets["frontend_manual_offset"].text())
        msg.frontend_manual_offset_control = self.fieldWidgets["frontend_manual_offset_control"].isChecked()
        msg.gain_level = self.fieldWidgets["gain_level"].currentData()

        self.configurationProvided.emit(msg)

    @QtCore.pyqtSlot(object)
    def handleReceivedConfiguration(self, msg:pipb.SamplingConfiguration):
        self.log.info("updating SamplingConfiguration from device")
        for f in self.intFields:
            self.fieldWidgets[f].setText(str(getattr(msg, f)))
        self.fieldWidgets["frontend_manual_offset"].setText(str(msg.frontend_manual_offset))
        self.fieldWidgets["frontend_manual_offset_control"].setChecked(msg.frontend_manual_offset_control)
        sel_idx = None
        gcb = self.fieldWidgets["gain_level"]
        for i in range(0, gcb.count()):
            if gcb.itemData(i) == msg.gain_level:
                sel_idx = i
                break
        if sel_idx is not None:
            gcb.setCurrentIndex(sel_idx)



class ThresholdConfigurationWidget(QtWidgets.QGroupBox):

    fieldWidgets: typing.Dict[str, QtWidgets.QLineEdit]

    configurationRequest = QtCore.pyqtSignal()
    measureNoiseFloorRequest = QtCore.pyqtSignal()
    configurationProvided = QtCore.pyqtSignal(object)
    statusReportProvided = QtCore.pyqtSignal(object)
    floatFields = ["sensitivity", "normalizedSensitivity", "noiseFloorMultiplier", "minNoiseFloor", "dynamicModeAlpha"]

    DETECTION_STATISTICS = pipb.ThresholdConfiguration.DetectionStatistic.items()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.log = logging.getLogger("thresholdcfg")
        self.setTitle("Threshold configuration")
        layout = QtWidgets.QFormLayout()
        self.fieldWidgets = {}
        def cw(name, title, def_val):
            w = QtWidgets.QLineEdit(def_val)
            layout.addRow(title, w)
            self.fieldWidgets[name] = w

        cw("sensitivity", "Sensitivity", "1.0")
        cw("normalizedSensitivity", "Dyn. sensitivity", "1.0")
        cw("noiseFloorMultiplier", "Noise floor mult.", "1.1")
        cw("minNoiseFloor", "Min. noise floor", "10.0")
        cw("dynamicModeAlpha", "Dynamic mode LPF Fc", "0.995")
        ce = QtWidgets.QCheckBox()
        layout.addRow("Compute expensive stats?", ce)
        self.fieldWidgets["compute_expensive_stats"] = ce

        def cl(name, title):
            w = QtWidgets.QLabel("")
            layout.addRow(title, w)
            self.fieldWidgets[name] = w

        cl("noise_floor", "Noise floor for operation")
        cl("measured_noise_floor", "Measured noise floor")
        cl("measured_mean_value", "Measured mean value")
        cl("measured_slope", "Measured slope")
        cl("measured_min_value", "Measured min value")

        sw = QtWidgets.QComboBox()
        for name, value in self.DETECTION_STATISTICS:
            sw.addItem(name, value)
        self.fieldWidgets["detectionStatistic"] = sw
        layout.addRow("Stat used for detection", sw)

        wc = QtWidgets.QPushButton("Write configuration")
        wc.released.connect(self.handleWriteConfiguration)
        rc = QtWidgets.QPushButton("Read configuration")
        rc.released.connect(lambda: self.configurationRequest.emit())
        layout.addRow(rc, wc)

        mnf = QtWidgets.QPushButton("Measure noise floor")
        mnf.released.connect(lambda: self.measureNoiseFloorRequest.emit())
        layout.addRow(mnf)
        self.setLayout(layout)

    @QtCore.pyqtSlot()
    def handleWriteConfiguration(self):
        self.log.info("writing configuration")

        msg = pipb.ThresholdConfiguration()
        for f in self.floatFields:
            setattr(msg, f, float(self.fieldWidgets[f].text()))

        msg.detectionStatistic = self.fieldWidgets["detectionStatistic"].currentData()
        msg.compute_expensive_stats = self.fieldWidgets["compute_expensive_stats"].isChecked()
        self.configurationProvided.emit(msg)

    @QtCore.pyqtSlot(object)
    def handleReceivedConfiguration(self, msg:pipb.ThresholdConfiguration):
        self.log.info("updating ThresholdConfiguration from device")
        for f in self.floatFields:
            self.fieldWidgets[f].setText(str(getattr(msg, f)))

        sel_idx = None
        gcb = self.fieldWidgets["detectionStatistic"]
        for i in range(0, gcb.count()):
            if gcb.itemData(i) == msg.detectionStatistic:
                sel_idx = i
                break
        if sel_idx is not None:
            gcb.setCurrentIndex(sel_idx)

        self.fieldWidgets["compute_expensive_stats"].setChecked(getattr(msg, "compute_expensive_stats", False))

    @QtCore.pyqtSlot(object)
    def handleStatusReport(self, msg:pipb.ThresholdStatusReport):
        self.fieldWidgets["noise_floor"].setText(str(msg.noiseFloor))
        self.fieldWidgets["measured_noise_floor"].setText(str(msg.measuredNoiseFloor))
        self.fieldWidgets["measured_mean_value"].setText(str(msg.meanValue))
        self.fieldWidgets["measured_slope"].setText(str(msg.meanValueSlope))
        self.fieldWidgets["measured_min_value"].setText(str(msg.minValue))


class SystemStatusWidget(QtWidgets.QGroupBox):
    fieldWidgets: typing.Dict[str, QtWidgets.QLabel]
    simulatedStateProvided = QtCore.pyqtSignal(object)
    rebootRequested = QtCore.pyqtSignal()
    restoreConfiguration = QtCore.pyqtSignal()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setTitle("System status")
        layout = QtWidgets.QFormLayout()
        self.fieldWidgets = {}
        for key, title in (("processed_sample_rate", "Processed sample rate"),
                           ("adc_read_sample_rate", "ADC read sample rate"),
                           ("battery_voltage", "Battery voltage"),
                           ("battery_soc", "Battery SoC"),
                           ("mcu_temperature_c", "MCU temperature"),
                           ("headphones_detected", "Headphones detected"),
                           ):
            self.fieldWidgets[key] = QtWidgets.QLabel()
            layout.addRow(title, self.fieldWidgets[key])

        self.fieldWidgets["sim_battery_voltage"] = QtWidgets.QLineEdit("3.7")
        self.fieldWidgets["sim_vibro_pwm_dc"] = QtWidgets.QLineEdit("")
        self.fieldWidgets["sim_vibro_pwm_dc"].setPlaceholderText("0.0-1.0")
        layout.addRow("Sim. batt. volt.", self.fieldWidgets["sim_battery_voltage"])
        layout.addRow("Sim. vibro. dc.", self.fieldWidgets["sim_vibro_pwm_dc"])
        pb = QtWidgets.QPushButton("Send sim. state")
        pb.released.connect(self.handleSimStateSend)
        layout.addRow(pb)
        pbReboot = QtWidgets.QPushButton("Reboot")
        pbReboot.released.connect(lambda: self.rebootRequested.emit())

        pbRestoreDefaultConf = QtWidgets.QPushButton("Restore default conf.")
        pbRestoreDefaultConf.released.connect(lambda: self.restoreConfiguration.emit())
        layout.addRow(pbReboot, pbRestoreDefaultConf)

        self.setLayout(layout)

    @QtCore.pyqtSlot(object)
    def handleReceivedStatus(self, msg: pipb.SystemStatus):
        for k in ["processed_sample_rate", "adc_read_sample_rate", "headphones_detected"]:
            self.fieldWidgets[k].setText(str(getattr(msg, k, "")))

        self.fieldWidgets["battery_voltage"].setText(f"{msg.battery_voltage:.2f} V")
        self.fieldWidgets["battery_soc"].setText(f"{msg.battery_soc} %")
        self.fieldWidgets["mcu_temperature_c"].setText(f"{msg.mcu_temperature_c:.1f} °C")

    @QtCore.pyqtSlot()
    def handleSimStateSend(self):
        msg = pipb.SetSimulatedStateRequest()
        batval = self.fieldWidgets["sim_battery_voltage"].text()
        if batval:
            msg.battery_voltage = float(batval)
            msg.has_battery_voltage = True
        else:
            msg.has_battery_voltage = False

        vibroval = self.fieldWidgets["sim_vibro_pwm_dc"].text()
        if vibroval:
            msg.vibro_pwm_dc = float(vibroval)
            msg.has_vibro_pwm_dc = True
        else:
            msg.has_vibro_pwm_dc = False

        self.simulatedStateProvided.emit(msg)