import logging

import pyqtgraph
from PyQt6 import QtWidgets, QtCore
import piapi
import piapiworker
import pi_pb2 as pipb
import sampleplot
import confwidgets

class MainWindow(QtWidgets.QMainWindow):
    log: logging.Logger
    api: piapi.PIAPI
    samplePlotWidget: sampleplot.CurrentSamplePlotWidget
    statsWidget: sampleplot.CurrentSampleStatsWidget
    samplingConfWidget: confwidgets.SamplingConfigurationWidget
    thresholdConfWidget: confwidgets.ThresholdConfigurationWidget
    systemStatusWidget: confwidgets.SystemStatusWidget

    def __init__(self, *args, **kwargs):
        port = kwargs.pop('port')
        super(MainWindow, self).__init__(*args, **kwargs)
        self.log = logging.getLogger("pigui")
        self.api = piapi.PIAPI(port)
        self.threadpool = QtCore.QThreadPool()
        self.apiworker = piapiworker.PIAPIWorker(self.api)
        self.apiworker.signals.messageReceived.connect(self.processMessage)
        self.threadpool.start(self.apiworker)
        self.api.request_identification()

        w = QtWidgets.QWidget()
        layout = QtWidgets.QHBoxLayout()
        self.createPlotWidgets(layout)
        w.setLayout(layout)
        self.setCentralWidget(w)

    def createPlotWidgets(self, layout):
        self.samplePlotWidget = sampleplot.CurrentSamplePlotWidget()
        self.apiworker.signals.detectionMeasurementReceived.connect(self.samplePlotWidget.handleDetectionMeasurement)
        self.apiworker.signals.referenceMeasurementReceived.connect(self.samplePlotWidget.handleReferenceMeasurement)

        self.statsWidget = sampleplot.CurrentSampleStatsWidget()
        self.apiworker.signals.detectionMeasurementReceived.connect(self.statsWidget.handleDetectionMeasurement)
        self.apiworker.signals.referenceMeasurementReceived.connect(self.statsWidget.handleReferenceMeasurement)

        self.samplingConfWidget = confwidgets.SamplingConfigurationWidget()
        self.samplingConfWidget.configurationProvided.connect(self.api.send_message)
        self.apiworker.signals.samplingConfigurationReceived.connect(self.samplingConfWidget.handleReceivedConfiguration)
        self.samplingConfWidget.configurationRequest.connect(lambda: self.api.send_message(pipb.SamplingConfigurationRequest()))
        self.samplingConfWidget.acquireReferenceSampleRequest.connect(lambda: self.api.send_message(pipb.PerformReferenceMeasurement()))
        self.samplingConfWidget.automaticOffsetAdjustmentRequest.connect(lambda: self.api.send_message(pipb.AutomaticOffsetAdjustmentRequest()))
        self.api.send_message(pipb.SamplingConfigurationRequest())


        self.thresholdConfWidget = confwidgets.ThresholdConfigurationWidget()
        self.thresholdConfWidget.configurationProvided.connect(self.api.send_message)
        self.apiworker.signals.thresholdConfigurationReceived.connect(self.thresholdConfWidget.handleReceivedConfiguration)
        self.thresholdConfWidget.configurationRequest.connect(lambda: self.api.send_message(pipb.ThresholdConfigurationRequest()))
        self.thresholdConfWidget.measureNoiseFloorRequest.connect(lambda: self.api.send_message(pipb.NoiseFloorMeasurementRequest()))
        self.apiworker.signals.thresholdStatusReportReceived.connect(self.thresholdConfWidget.handleStatusReport)
        self.api.send_message(pipb.ThresholdConfigurationRequest())

        self.systemStatusWidget = confwidgets.SystemStatusWidget()
        self.systemStatusWidget.simulatedStateProvided.connect(lambda x: self.api.send_message(x))
        self.systemStatusWidget.rebootRequested.connect(lambda: self.api.send_message(pipb.BoardRebootRequest()))
        self.systemStatusWidget.restoreConfiguration.connect(lambda: self.api.send_message(pipb.RestoreDefaultConfigurationRequest()))
        self.apiworker.signals.systemStatusReceived.connect(self.systemStatusWidget.handleReceivedStatus)

        leftLayout = QtWidgets.QVBoxLayout()
        leftLayout.addWidget(self.samplePlotWidget)
        confLayout = QtWidgets.QHBoxLayout()

        confLayout.addWidget(self.samplingConfWidget)
        confLayout.addWidget(self.thresholdConfWidget)
        confLayout.addWidget(self.systemStatusWidget)
        leftLayout.addLayout(confLayout)

        layout.addLayout(leftLayout)
        statsLayout = QtWidgets.QVBoxLayout()
        statsLayout.addWidget(self.statsWidget)
        btn = QtWidgets.QPushButton("Clear stats")
        btn.released.connect(self.statsWidget.clearSamples)
        statsLayout.addWidget(btn)
        layout.addLayout(statsLayout)

    def stopApiWorker(self):
        self.apiworker.stop()
        self.threadpool.waitForDone(1000)

    def closeEvent(self, evt):
        self.log.info("closeEvent()")
        if not self.apiworker.should_stop:
            self.stopApiWorker()
        super().closeEvent(evt)

    def processMessage(self, msg):
        """Process incoming message from device"""
        if type(msg) == pipb.BoardIdentification:
            self.log.info("BoardIdentification: %s" % msg)
