from google.protobuf.message import Message
import enum
import pi_pb2 as pipb
import nanopb_pb2
import struct

# gets a list of message types
MESSAGE_TYPES = [v for v in vars(pipb).values() if isinstance(v, type) and issubclass(v, Message)]
TYPE_BY_MSGID = {}

def msgid(m):
    return m.DESCRIPTOR.GetOptions().Extensions[nanopb_pb2.nanopb_msgopt].msgid

for m in MESSAGE_TYPES:
    TYPE_BY_MSGID[msgid(m)] = m


class Nanoparser:
    """Implements protobuf message framing, serialization and deserialization according to nanoparser.h"""
    MAGIC1 = 0xFE
    MAGIC2 = 0xFD
    CHECKSUM_INIT = 0x42

    def __init__(self):
        self._state = self.State.IDLE
        self._payload_ba = bytearray()
        self._expected_payload_len = 0
        self._length_ba = bytearray()
        self.msgid = None
        self.recv_checksum = self.CHECKSUM_INIT

    def parse_byte(self, b) -> tuple:
        res = self.Result.NONE
        msg = None
        if self._state == self.State.IDLE and b == self.MAGIC1:
            self._state = self.State.GOT_MAGIC1
            self._payload_ba = bytearray()
            self._length_ba = bytearray()
            self.msgid = None
            self._expected_payload_len = 0
            self.recv_checksum = self.CHECKSUM_INIT ^ b

        elif self._state == self._state.GOT_MAGIC1 and b == self.MAGIC2:
            self._state = self.State.GOT_MAGIC2
            self.recv_checksum ^= b
        elif self._state == self.State.GOT_MAGIC2:
            self.msgid = int(b)
            self._state = self.State.GOT_MSGID
            self.recv_checksum ^= b
        elif self._state == self.State.GOT_MSGID:
            self._length_ba.append(b)
            self.recv_checksum ^= b
            self._state = self.State.GOT_LENGTH1
        elif self._state == self.State.GOT_LENGTH1:
            self._length_ba.append(b)
            self.recv_checksum ^= b
            self._state = self.State.GOT_LENGTH2
            self._expected_payload_len = struct.unpack("<h", self._length_ba)[0]
            if self._expected_payload_len > 16384:
                res = self.Result.INVALID_LENGTH
                self._state = self.State.IDLE
            elif self._expected_payload_len == 0:
                self._state = self.State.GOT_PAYLOAD
        elif self._state == self.State.GOT_LENGTH2 and len(self._payload_ba) < self._expected_payload_len:
            self._payload_ba.append(b)
            self.recv_checksum ^= b
            if len(self._payload_ba) == self._expected_payload_len:
                self._state = self.State.GOT_PAYLOAD
        elif self._state == self.State.GOT_PAYLOAD:
            if b != self.recv_checksum:
                res = self.Result.INVALID_CRC
            else:
                msgclass = TYPE_BY_MSGID[self.msgid]
                msg = msgclass()
                msg.ParseFromString(bytes(self._payload_ba))
                res = self.Result.MSG_OK
            self._state = self.State.IDLE

        return res, msg

    class State(enum.Enum):
        IDLE = enum.auto()
        GOT_MAGIC1 = enum.auto()
        GOT_MAGIC2 = enum.auto()
        GOT_MSGID = enum.auto()
        GOT_LENGTH1 = enum.auto()
        GOT_LENGTH2 = enum.auto()
        GOT_PAYLOAD = enum.auto()

    class Result(enum.Enum):
        NONE = enum.auto()
        MSG_OK = enum.auto()
        INVALID_LENGTH = enum.auto()
        INVALID_CRC = enum.auto()

    @staticmethod
    def serialize_msg(msg):
        ba = bytearray([Nanoparser.MAGIC1, Nanoparser.MAGIC2])    # header
        ba.append(msgid(msg))           # message id
        msg_bytes = msg.SerializeToString()
        ba.extend(struct.pack("<h", len(msg_bytes)))
        ba.extend(msg_bytes)
        csum = Nanoparser.CHECKSUM_INIT
        for b in ba:
            csum ^= b
        ba.append(csum)
        return ba


if __name__ == "__main__":

    m = tpb.LogEntry()
    m.message = "test 123"
    msg_serialized = Nanoparser.serialize_msg(m)
    print("serialized msg=%s" % msg_serialized)
    print("decoding the message...")

    p = Nanoparser()
    for b in msg_serialized:
        res, msg = p.parse_byte(b)
        if res == Nanoparser.Result.MSG_OK:
            print("message decoded, msg=%s" % msg)
        elif res != Nanoparser.Result.NONE:
            print("result = %s" % res)


