import logging
import serial
from numpy.f2py.auxfuncs import throw_error

import nanoparser
import pi_pb2 as pipb
import typing
import time
import socket

try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

class TCPSocketDev:
    
    def __init__(self, log:logging.Logger, connstr:str, timeout:float):
        self.log = log
        proto, addr, port = connstr.split(":")
        port = int(port)
        self.log.info(f"connecting to tcp:{addr}:{port}")
 
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM) 
        s.setblocking(False)
        s.settimeout(timeout)
        s.connect((addr, port))
        self.socket = s

    def read(self, max_num_bytes):
        try:
            return self.socket.recv(max_num_bytes)
        except TimeoutError:
            return []


    def write(self, data):
        return self.socket.send(data)


class DataThroughputMonitor:
    def __init__(self, log_period_s: float, log: logging.Logger):
        self.log = log
        self.log_period_s = log_period_s
        self.start_time = time.time()
        self.bytes_received = 0
        self.messages_received = 0

    def reset(self):
        self.start_time = time.time()
        self.bytes_received = 0
        self.messages_received = 0

    def add_received_data(self, data_len: int):
        self.bytes_received += data_len
        if (time.time() - self.start_time) > self.log_period_s:
            self.log_throughput()
            self.reset()
    def add_received_message(self, num_messages: int = 1):
        self.messages_received += num_messages

    def log_throughput(self):
        dt = time.time() - self.start_time
        throughput_bytes_s = self.bytes_received / dt
        msg_s = self.messages_received / dt
        self.log.debug(f"Throughput {throughput_bytes_s:.1f} byte/sec, {msg_s:.1f}msg/sec")


class PIAPI:
    parser: nanoparser.Nanoparser
    log: logging.Logger
    dev: serial.Serial


    def __init__(self, port: str):
        self.log = logging.getLogger("piapi")
        self.recv_throughput_monitor = DataThroughputMonitor(5.0, self.log)
        self.parser = nanoparser.Nanoparser()
        self.port = port
        self._open_port()
        self.log.info("initialized on port %s" % port)

    def _open_port(self):
        timeout = 0.01
        if self.port.startswith("tcp:"):
           self.dev = TCPSocketDev(self.log, self.port, timeout)
        else:
            if "," in self.port:
                ports = self.port.split(",")
            else:
                ports = [self.port]
            for port in ports:
                self.log.info(f"connecting to serial:{port}")
                self.dev = serial.Serial(port, timeout=timeout)

    def reconnect(self):
        if self.dev:
            self.dev.close()
            self.dev = None
        while True:
            time.sleep(1)
            self.log.info("Reconnection attempt...")
            try:
                self._open_port()
                return
            except serial.SerialException:
                pass


    def _read(self, max_num_bytes=128) -> typing.Optional[list]:
        if self.dev is None:
            return []
        try:
            data_read = self.dev.read(max_num_bytes)
            self.recv_throughput_monitor.add_received_data(len(data_read))
            return data_read

        except serial.SerialException:
            self.log.warning("Serial connection error, re-connecting...")
            self.reconnect()
            return []

    def _write(self, data):
        if self.dev:
            return self.dev.write(data)
        else:
            return 0

    def send_message(self, msg) -> bool:
        return self._write(self.parser.serialize_msg(msg)) > 0

    def receive_messages(self, filter=None, timeout=1, abort_early=True) -> list:
        """
        Receives message(s) from the device.

        :param filter: types of messages that are returned: either a single message type
                        (e.g. pipb.LogEntry), or a list of types
        :param timeout: timeout in seconds
        :return: list of messages that have been received (which can be empty list)
        """
        ret = []
        # filter can be either a single type or a list of types
        if filter and not isinstance(filter, Iterable):
            filter = [filter]
        abort_receive_loop = False

        end_time = time.time() + timeout
        while time.time() < end_time and not abort_receive_loop:
            bytes_recv = self._read(128)
            for b in bytes_recv:
                res, msg = self.parser.parse_byte(b)
                if res == self.parser.Result.MSG_OK:
                    self.recv_throughput_monitor.add_received_message(1)
                    self._is_connected = True
                    if type(msg) == pipb.LogEntry:
                        self._handle_log_entry(msg)
                    else:
                        # no need to log LogEntry, as it's being logged already
                        self.log.log(logging.TRACE, "received message: %s" % msg.__class__.__name__)
                    if filter:
                        if type(msg) in filter:
                            ret.append(msg)
                            if abort_early:
                                abort_receive_loop = True
                    elif type(msg) != pipb.LogEntry:
                        # no filter
                        ret.append(msg)
                        if abort_early:
                            abort_receive_loop = True
                elif res == self.parser.Result.INVALID_CRC:
                    self.log.warning("received data with invalid crc")
                elif res == self.parser.Result.INVALID_LENGTH:
                    self.log.warning("received data with invalid crc")
        return ret

    def _handle_log_entry(self, msg):
        self.log.info("D: %s" % msg.message)

    def request_identification(self):
        self.send_message(pipb.BoardIdentificationRequest())



if __name__ == "__main__":
    import sys
    logging.basicConfig(level=logging.DEBUG)
    log = logging.getLogger("main")

    if len(sys.argv) < 2:
        log.error("Serial port not provided")
        sys.exit(1)

    api = PIAPI(sys.argv[1])
    api.request_identification()
    while True:
        print(api.receive_messages(timeout=1))
        time.sleep(0.1)