from PyQt6 import QtWidgets, QtCore
import pyqtgraph as pg
import pi_pb2 as pipb
import typing
import numpy as np
import logging

class CurrentSamplePlotWidget(QtWidgets.QWidget):
    plotItems: typing.List[typing.Tuple[float, pipb.DetectionMeasurement]]
    refPlotItem: object
    log: logging.Logger
    latestTimestampUs: float
    plotItemDisplayTimeout = 0.0

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.log = logging.getLogger("sampleplot")

        pg.setConfigOption("useOpenGL", True)
        self.graphWidget = pg.PlotWidget()
        self.graphWidget.setLabel('bottom', 'Time (us)')
        self.graphWidget.setLabel('left', 'Signal amplitude')
        self.graphWidget.setYRange(0, 2**14-1)
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.graphWidget)
        self.setLayout(layout)
        self.plotItems = []
        self.refPlotItem = None
        self.latestTimestampUs = 0.0
        self.numMeasurements = 0

    @QtCore.pyqtSlot(object)
    def handleDetectionMeasurement(self, msg: pipb.DetectionMeasurement):
        self.log.log(logging.TRACE, "processing measurement %d" % msg.timestamp_us)
        time_us = list(range(0, len(msg.sample.adc_sample), 1))
        adc_samples = list(np.array(msg.sample.adc_sample))

        if self.plotItemDisplayTimeout > 0.0:
            pi = self.graphWidget.plot(time_us, adc_samples)
            self.plotItems.append((msg.timestamp_us, pi))
            self.latestTimestampUs = msg.timestamp_us
            self.handlePlotItemsCleanup()
        else:
            if len(self.plotItems) > 0:
                self.graphWidget.removeItem(self.plotItems[0][1])
                self.plotItems.clear()

            pi = self.graphWidget.plot(time_us, adc_samples)
            self.plotItems.append((msg.timestamp_us, pi))


    @QtCore.pyqtSlot(object)
    def handleReferenceMeasurement(self, msg: pipb.ReferenceMeasurement):
        if self.refPlotItem:
            self.graphWidget.removeItem(self.refPlotItem)
        time_us = list(range(0, len(msg.sample.adc_sample) , 1))
        adc_samples = list(np.array(msg.sample.adc_sample))
        pi = self.graphWidget.plot(time_us, adc_samples, pen=pg.mkPen(color=(255, 0, 0), width=3))
        self.refPlotItem = pi

    def handlePlotItemsCleanup(self):
        for itm in self.plotItems:
            timestamp, pi = itm
            dt = (self.latestTimestampUs - timestamp) * 1e-6
            if dt > self.plotItemDisplayTimeout:
                # remove it
                self.graphWidget.removeItem(pi)
                self.plotItems.remove(itm)
            else:
                # decrease transparency
                alpha = 1-dt/self.plotItemDisplayTimeout
                alpha = (10**alpha - 1.0) / 9.0
                pi.setPen(alpha)

class CurrentSampleStatsWidget(QtWidgets.QWidget):
    #stats = ["mean_amp", "lin_fit_slope", "lin_fit_offset", "exp_fit_exponent", "exp_fit_intercept", "integral", "amplitude", "weighted_delta_ref"]
    stats = ["mean_amp", "mean_amp_filtered", "integral", "weighted_delta_ref", "amplitude"]
    stat_types = ["min", "max", "mean", "p-p", "std"]
    plotWidgets: typing.Dict[str, pg.PlotItem]
    numWidgets: typing.Dict[str, typing.Dict[str, QtWidgets.QLabel]]
    glw: pg.GraphicsLayoutWidget
    log: logging.Logger
    samples: typing.List[typing.Tuple[float, pipb.DetectionMeasurement]]
    latestTimestampUs: float
    plotDisplayTimeout = 30
    deltaRefCb: QtWidgets.QCheckBox
    referenceMeasurement: typing.Optional[pipb.ReferenceMeasurement]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.log = logging.getLogger("statsplot")
        mainLayout = QtWidgets.QHBoxLayout()
        numLayout = QtWidgets.QVBoxLayout()
        self.glw = pg.GraphicsLayoutWidget()
        self.plotWidgets = {}
        self.numWidgets = {}
        self.referenceMeasurement = None
        for stat in self.stats:
            self.glw.nextRow()
            pw = self.glw.addPlot()
            pw.setTitle(stat)

            self.plotWidgets[stat] = pw
            self.numWidgets[stat] = {}
            nl = QtWidgets.QVBoxLayout()
            for st in self.stat_types:
                nw = QtWidgets.QLabel()
                nl.addWidget(nw)
                self.numWidgets[stat][st] = nw

            nl.addItem(QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding))
            numLayout.addLayout(nl)

        drl = QtWidgets.QHBoxLayout()
        drl.addWidget(QtWidgets.QLabel("Delta ref?"))
        self.deltaRefCb = QtWidgets.QCheckBox()
        drl.addWidget(self.deltaRefCb)
        numLayout.addLayout(drl)

        mainLayout.addWidget(self.glw)
        mainLayout.addLayout(numLayout)
        self.setLayout(mainLayout)
        self.samples = []
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.updatePlot)
        self.timer.start(200)
        self.latestTimestampUs = 0

    @QtCore.pyqtSlot(object)
    def handleDetectionMeasurement(self, msg: pipb.DetectionMeasurement):
        self.latestTimestampUs = msg.timestamp_us
        self.samples.append((msg.timestamp_us, msg))

    @QtCore.pyqtSlot(object)
    def handleReferenceMeasurement(self, msg: pipb.ReferenceMeasurement):
        self.referenceMeasurement = msg

    @QtCore.pyqtSlot()
    def updatePlot(self):
        self.handleSamplesCleanup()
        timestamps = []
        stats_vals = {}
        ref_vals = {}

        for stat in self.stats:
            stats_vals[stat] = []

        for ts, msg in self.samples:
            timestamps.append(ts * 1e-6)
            for stat in self.stats:
                if stat in ("amplitude", "phase"):
                    stats_vals[stat].append(getattr(msg, stat))
                else:
                    stats_vals[stat].append(getattr(msg.sample, stat))


        if len(timestamps) > 0:
            for stat in self.stats:
                arr = np.array(stats_vals[stat])
                if self.deltaRefCb.isChecked() and self.referenceMeasurement and stat not in ("amplitude", "phase"):
                    arr = arr - getattr(self.referenceMeasurement.sample, stat)

                self.plotWidgets[stat].plot(timestamps, list(arr), clear=True)
                if not self.deltaRefCb.isChecked() and self.referenceMeasurement and stat not in ("amplitude", "phase"):
                    rv = getattr(self.referenceMeasurement.sample, stat)
                    self.plotWidgets[stat].plot(timestamps, [rv] * len(timestamps), pen=(255, 0, 0))
                stat_types = self.stat_types
                stat_vals = [
                    np.min(arr), np.max(arr), np.mean(arr), np.ptp(arr), np.std(arr)
                ]
                for stat_name, val in zip(stat_types, stat_vals):
                    self.numWidgets[stat][stat_name].setText("%s = %6.3f" % (stat_name, val))

    def handleSamplesCleanup(self):
        for itm in self.samples:
            timestamp, msg = itm
            dt = (self.latestTimestampUs - timestamp) * 1e-6
            if dt > self.plotDisplayTimeout:
                self.samples.remove(itm)

    def clearSamples(self):
        self.samples = []