#include "nanoparser.h"
#include <pb_encode.h>

#define MAGIC_HEADER1 0xFE
#define MAGIC_HEADER2 0xFD
#define MAGIC_CHECKSUM_INIT 0x42

void parser_init_msg(parser_msg_t* msg)
{
    msg->_state = PARSER_STATE_IDLE;
}

parser_result_t parser_parse_byte(parser_msg_t* msg, uint8_t byte)
{
    parser_result_t ret = PARSER_RESULT_NONE;
    switch (msg->_state)
    {
        case PARSER_STATE_IDLE:
            if (byte == MAGIC_HEADER1) {
                msg->_state = PARSER_STATE_GOT_MAGIC1;
                msg->_recv_checksum = MAGIC_CHECKSUM_INIT ^ byte;
                msg->_pidx = 0;
            }
            break;
        case PARSER_STATE_GOT_MAGIC1:
            if (byte == MAGIC_HEADER2) {
                msg->_state = PARSER_STATE_GOT_MAGIC2;
                msg->_recv_checksum ^= byte;
            } else {
                //reset statemachine
                msg->_state = PARSER_STATE_IDLE;
            }
            break;
        case PARSER_STATE_GOT_MAGIC2:
            msg->msgid = byte;
            msg->_state = PARSER_STATE_GOT_MSGID;
            msg->_recv_checksum ^= byte;
            break;
        case PARSER_STATE_GOT_MSGID:
            msg->length = byte;
            msg->_recv_checksum ^= byte;
            msg->_state = PARSER_STATE_GOT_LENGTH1;
            break;
        case PARSER_STATE_GOT_LENGTH1:
            msg->length |= ((uint16_t)byte) << 8;
            msg->_recv_checksum ^= byte;

            if (msg->length == 0)
                msg->_state = PARSER_STATE_GOT_PAYLOAD;
            else
                msg->_state = PARSER_STATE_GOT_LENGTH2;

            if (msg->length > PARSER_MAX_MESSAGE_LENGTH)
            {
                ret = PARSER_RESULT_INVALID_LENGTH;
            }
            break;
        case PARSER_STATE_GOT_LENGTH2:
            msg->payload[msg->_pidx] = byte;
            msg->_pidx++;
            msg->_recv_checksum ^= byte;
            if (msg->_pidx >= msg->length) {
                msg->_state = PARSER_STATE_GOT_PAYLOAD;
            }
            break;
        case PARSER_STATE_GOT_PAYLOAD:
            msg->checksum = byte;
            if (msg->checksum == msg->_recv_checksum)
            {
                //hooray, received a valid checksum
                ret = PARSER_RESULT_MSG_OK;
            } else {
                ret = PARSER_RESULT_INVALID_CRC;
            }
            //prepare message struct for repeated parsing
            msg->_state = PARSER_STATE_IDLE;
            break;
    }
    return ret;
}

parser_result_t
parser_frame_message(const void *src_struct, const pb_msgdesc_t *fields, uint8_t msgid, uint8_t *buf, size_t bufsize,
                     size_t *outbufsize) {

    //zero the filled buffer size in order to not to deal with it at every error condition
    *outbufsize = 0;
    size_t _outsize = 0;
    //6 bytes is the minimum frame overhead with 0 payload
    if (bufsize <  6) {
        return PARSER_RESULT_INVALID_LENGTH;
    }

    uint8_t csum = MAGIC_CHECKSUM_INIT;
    buf[0] = MAGIC_HEADER1;
    buf[1] = MAGIC_HEADER2;
    buf[2] = msgid;
    //buf[3], buf[4] -> length
    //create output stream for the payload section of framed message
    pb_ostream_t os = pb_ostream_from_buffer(buf + 5, bufsize -  6);
    if (!pb_encode(&os, fields, src_struct)) {
        return PARSER_RESULT_ENCODE_FAILED;
    }

    if (os.bytes_written > UINT16_MAX)
        return PARSER_RESULT_INVALID_LENGTH;
    uint16_t plen = os.bytes_written;

    buf[3] = plen & (uint8_t)0xFF;
    buf[4] = plen >> (uint8_t)8;
    _outsize = 5 + os.bytes_written;
    for (size_t i=0; i < _outsize; i++)
    {
        csum ^= buf[i];
    }
    buf[_outsize] = csum;
    _outsize++;

    *outbufsize = _outsize;

    return PARSER_RESULT_MSG_OK;
}


