#ifndef LIB_NANOPARSER_H
#define LIB_NANOPARSER_H

#include <stdint.h>
#include <pb_common.h>

#ifdef __cplusplus
extern "C" {
#endif


/**
 * Nanoparser is framing protocol implementation suitable for wrapping certain kind of
 * binary payloads for streaming transport layers (e.g. serial ports, TCP packets).
 *
 * Nanoparser protocol format:
 * Field name:length in bytes
 *
 * MAGIC_HEADER:2 MSGID:1 LENGTH:2 PAYLOAD:L CHECKSUM:1
 *
 * MAGIC_HEADER is chosen to be {0xFE, 0xFD}
 *
 * CHECKSUM is a single byte checksum, calculated on the whole packet starting from
 * MAGIC_HEADER until the last PAYLOAD byte. It is calculated by sequential XOR-ing of
 * every byte in the packet with previous byte (first byte is XOR-ed with constant 0x42).
 *
 * The checksum is deliberately very simple one, as it's main purpose is to guard against
 * accidental decoding of garbage, if the MAGIC_HEADER is found in data.
 *
 * Algorithm:
 *
 * uint8_t c = 0x42;
 * for (int i=0; i < buflen; i++)
 *     c = c ^ buffer[i];
 *
 * variable c contains checksum
 *
 */

#define PARSER_MAX_MESSAGE_LENGTH (16 * 1024)

typedef enum {
    PARSER_STATE_IDLE = 0,
    PARSER_STATE_GOT_MAGIC1,
    PARSER_STATE_GOT_MAGIC2,
    PARSER_STATE_GOT_MSGID,
    PARSER_STATE_GOT_LENGTH1,
    PARSER_STATE_GOT_LENGTH2,
    PARSER_STATE_GOT_PAYLOAD,
} parser_state_t;

/**
 * Static message structure
 */
typedef struct {
    uint8_t msgid;
    uint16_t length;
    uint8_t payload[PARSER_MAX_MESSAGE_LENGTH];
    uint8_t checksum;

    uint8_t _recv_checksum;
    uint16_t _pidx;
    parser_state_t _state;
} parser_msg_t;

/**
 * Result type for parsing operations
 */
typedef enum {
    PARSER_RESULT_NONE = 0, // no result, waiting for more data
    PARSER_RESULT_MSG_OK, // message was successfully received
    PARSER_RESULT_INVALID_LENGTH, // message length exceeds maximum length or is zero
    PARSER_RESULT_INVALID_CRC, // invalid CRC, statemachine has been reset
    PARSER_RESULT_ENCODE_FAILED, // invalid CRC, statemachine has been reset
} parser_result_t;

/**
 * Initializes message for reception
 * @param msg
 */
void parser_init_msg(parser_msg_t *msg);

/**
 * Executes parser statemachine on a single byte at a time.
 * Caller is expected to have initialized the msg beforehand, using parser_init_msg()
 *
 * @param msg pointer to message structure
 * @param byte incoming byte to parser
 * @return parser_result_t
 */
parser_result_t parser_parse_byte(parser_msg_t *msg, uint8_t byte);

/**
 * Encodes and serializes a given message to buffer- adds header data, calculates checksum etc.
 *
 * @param src_struct message struct
 * @param fields message fields
 * @param msgid message identifier
 * @param buf buffer to which message is serialized
 * @param bufsize size of buf
 * @param outbufsize pointer to serialized message size
 * @return on success, PARSER_RESULT_MSG_OK is returned.
 */
parser_result_t
parser_frame_message(const void *src_struct, const pb_msgdesc_t *fields, uint8_t msgid, uint8_t *buf, size_t bufsize,
                     size_t *outbufsize);

/**
 * Define which decreases the likelihood of mixing up message type and id
 */
#define PARSER_FRAME_MESSAGE(src, typ, buf, bufsize, outbufsize) parser_frame_message(src, typ##_fields, typ##_msgid, buf, bufsize, outbufsize)

#ifdef __cplusplus
}
#endif

#endif //LIB_NANOPARSER_H