//
// Created by R on 23.18.2.
//

#include "PISampler.h"
#include <pico/stdlib.h>
#include <hardware/adc.h>
#include <hardware/pwm.h>
#include <hardware/clocks.h>
#include <hardware/dma.h>
#include <hardware/irq.h>
#include <cstring>
#include <algorithm>
#include <cmath>
#include <hardware/structs/syscfg.h>
#include <hardware/exception.h>
#include "off_pulse_sm0.pio.h"
#include "on_pulse_sm1.pio.h"
#include "sample_counter_sm2.pio.h"

static PISampler* PI_INST=nullptr;

void adc_dma_finished_cb() {
    PI_INST->adcDMAFinishedCb();
}

PISampler::PISampler(ADC& adc) : _sampling_cfg(SamplingConfiguration_init_default), _adc(adc),
                                 _frontend_offset_duty_cycle(0.0f), _sampleCounter(0),
                                 _sampleHistoryBufClearReq(false), _samplingConfigurationChanged(false),
                                 _numAdcDecayCurvesRead(0)
{
    _sampling_cfg.pulse_len_us = 90;
    _sampling_cfg.frontend_manual_offset = 0.5f;
    _sampling_cfg.frontend_manual_offset_control = false;

    _sampling_cfg.num_oversample_measurement = 16;
    _sampling_cfg.num_oversample_reference = MAX_SAMPLE_HISTORY;
    _sampling_cfg.num_samples_analyze = 44;
    _sampling_cfg.num_samples_skip = 18;
    _sampling_cfg.pulse_period_us = 1400;
    _sampling_cfg.num_undersample_processing = 1;
    _sampling_cfg.gain_level = SamplingConfiguration_FrontendGainLevel_GAIN0;
    _sampling_cfg.ref_sample_update_period_us = 800 * 1000;
    _sampling_cfg.auto_offset_adjustment_range_min = 12000;
    _sampling_cfg.auto_offset_adjustment_range_max = 13000;
    PI_INST = this;
}

bool PISampler::init() {
    _debugInit();
    _pwm_init();
    gpio_init(PIN_GAIN_10K);
    gpio_set_dir(PIN_GAIN_10K, true);
    gpio_put(PIN_GAIN_10K, false);

    gpio_init(PIN_GAIN_4_22K);
    gpio_set_dir(PIN_GAIN_4_22K, true);
    gpio_put(PIN_GAIN_4_22K, false);

    queue_init(&_rawSamplesQueue, sizeof(RawSample), 8);

    //setupIRQSampling();
    setupPIOSampling();
    _adc.register_dma_finished_callback(adc_dma_finished_cb);
    _adc.acquire_samples_dma_start(_currentADCSample.sample,  ADC_SAMPLE_DEPTH);
    return true;
}

bool PISampler::_pwm_init() {
    _offset_pwm_slice = pwm_gpio_to_slice_num(PIN_OFFSET_PWM);
    _offset_pwm_chan = pwm_gpio_to_channel(PIN_OFFSET_PWM);

    _offset_pwm_period_ticks = clock_get_hz(clk_sys) / OFFSET_PWM_FREQ_HZ;
    pwm_set_wrap(_offset_pwm_slice, _offset_pwm_period_ticks);
    set_offset_duty_cycle(0.5f, false);
    pwm_set_enabled(_offset_pwm_slice, true);
    gpio_init(PIN_OFFSET_PWM);
    gpio_set_function(PIN_OFFSET_PWM, GPIO_FUNC_PWM);

    return true;
}

void PISampler::set_offset_duty_cycle(float dc, bool manual) {
    dc = std::clamp(dc, 0.0f, 1.0f);
    _frontend_offset_duty_cycle = dc;
    _sampling_cfg.frontend_manual_offset = dc;
    _sampling_cfg.frontend_manual_offset_control = false;
    uint16_t level = (uint16_t)(_offset_pwm_period_ticks - 1) * dc;
    pwm_set_chan_level(_offset_pwm_slice, _offset_pwm_chan, level);
}


bool PISampler::set_sampling_configuration(const SamplingConfiguration &cfg) {
    if (cfg.num_samples_skip + cfg.num_samples_analyze > ADC_SAMPLE_DEPTH)
        return false;
    if (cfg.pulse_len_us > 500 * 1000)
        return false;
    if (cfg.num_samples_analyze == 0)
        return false;
    if (cfg.num_oversample_measurement == 0 ||  cfg.num_oversample_measurement > MAX_SAMPLE_HISTORY)
        return false;
    if (cfg.num_oversample_reference == 0 ||  cfg.num_oversample_reference > MAX_SAMPLE_HISTORY)
        return false;
    if (cfg.frontend_manual_offset < 0.0f || cfg.frontend_manual_offset > 1.0f)
        return false;

    if (cfg.frontend_manual_offset_control)
        set_offset_duty_cycle(cfg.frontend_manual_offset, true);

    _sampling_cfg = cfg;

    if (_sampling_cfg.num_undersample_processing == 0)
        _sampling_cfg.num_undersample_processing = 1;


    _set_frontend_gain(_sampling_cfg.gain_level);
    //update PIO settings- pulse length & period
    checkSamplingConfigurationForPIO();
    updatePIOSamplingConfiguration();
    _samplingConfigurationChanged = true;
    return true;
}

void PISampler::adcDMAFinishedCb() {
    _adc.acquire_samples_dma_start(_currentADCSample.sample,  ADC_SAMPLE_DEPTH);

    _sampleCounter++;
    if (_debugEnabled)
        gpio_put(DBG1_PIN, true);
    _numAdcDecayCurvesRead++;
    /*
    if (_acquireRawSamples && !queue_is_full(&_rawSamplesQueue)) {
        RawSample rs = RawSample_init_default;
        rs.timestamp_us = time_us_64();
        assert(ADC_SAMPLE_DEPTH <= (sizeof(rs.adc_sample) / sizeof(rs.adc_sample[0])));
        for (int i=0; i < ADC_SAMPLE_DEPTH; i++) {
            rs.adc_sample[i] = _currentADCSample.sample[i];
        }
        rs.adc_sample_count = ADC_SAMPLE_DEPTH;
        if (!queue_try_add(&_rawSamplesQueue, &rs))
            gpio_put(DBG1_PIN, true);
    }

    */

    if (_sampleHistoryBufClearReq) {
        _sampleHistoryBuf.clear();
        _sampleHistoryBufClearReq = false;
    }

    _sampleHistoryBuf.add(_currentADCSample);
    bool process_sample = _sampling_cfg.num_undersample_processing && (_sampleCounter % _sampling_cfg.num_undersample_processing) == 0;

    if (process_sample && _sampleHistoryBuf.size() >= _sampling_cfg.num_oversample_measurement) {
        uint32_t sample_history_sum[ADC_SAMPLE_DEPTH] = {0};
        for (int i=0; i < _sampling_cfg.num_oversample_measurement; i++) {
            for (int j=0; j < ADC_SAMPLE_DEPTH; j++) {
                sample_history_sum[j] += _sampleHistoryBuf.at(i).sample[j];
            }
        }
        decay_sample_t si;
        si.num_samples = _sampling_cfg.num_samples_analyze;
        for (int j=0; j < si.num_samples; j++) {
            si.sample[j] = (float) sample_history_sum[j + _sampling_cfg.num_samples_skip] /
                                  _sampling_cfg.num_oversample_measurement;
        }
        si.timestamp_us = time_us_64();

        auto added = decaySampleQueue.try_add(si);
        if (added && _debugEnabled) {
            gpio_put(DBG2_PIN, true);
            busy_wait_us(1);
            gpio_put(DBG2_PIN, false);
        }
    }
    if (_debugEnabled)
        gpio_put(DBG1_PIN, false);
}

void PISampler::_debugInit() {
    _debugEnabled = true;
    if (_debugEnabled) {
        gpio_init(DBG1_PIN);
        gpio_set_dir(DBG1_PIN, true);

        gpio_init(DBG2_PIN);
        gpio_set_dir(DBG2_PIN, true);
    }
}

bool PISampler::_set_frontend_gain(const SamplingConfiguration_FrontendGainLevel &gain) {
    switch (gain) {
        case SamplingConfiguration_FrontendGainLevel_GAIN0:
            gpio_put(PIN_GAIN_10K, false);
            gpio_put(PIN_GAIN_4_22K, false);
            break;
        case SamplingConfiguration_FrontendGainLevel_GAIN1:
            gpio_put(PIN_GAIN_10K, true);
            gpio_put(PIN_GAIN_4_22K, false);
            break;
        case SamplingConfiguration_FrontendGainLevel_GAIN2:
            gpio_put(PIN_GAIN_10K, false);
            gpio_put(PIN_GAIN_4_22K, true);
            break;
        case SamplingConfiguration_FrontendGainLevel_GAIN3:
            gpio_put(PIN_GAIN_10K, true);
            gpio_put(PIN_GAIN_4_22K, true);
            break;
        default:
            gpio_put(PIN_GAIN_10K, false);
            gpio_put(PIN_GAIN_4_22K, false);
    }
    return true;
}

bool PISampler::samplingConfigurationChanged() {
    auto tmp = _samplingConfigurationChanged;
    _samplingConfigurationChanged = false;
    return tmp;
}

void PISampler::setupPIOSampling() {
    checkSamplingConfigurationForPIO();
    auto pio = pio0;
    _pio = pio;
    {
        _pulse_period_sm = pio_claim_unused_sm(pio, true);
        auto sm0_freq = 1.0f / (5 * 1e-6f);
        float clkdiv = clock_get_hz(clk_sys) / sm0_freq;

        auto sm0_prog_offs = pio_add_program(pio, &off_pulse_sm0_program);
        pio_off_pulse_sm0_init(pio, _pulse_period_sm, sm0_prog_offs, clkdiv, _sampling_cfg.pulse_period_us / 100);
    }
    {
        _pulse_length_sm = pio_claim_unused_sm(pio, true);
        auto sm1_freq = 1.0f / (10 * 1e-6f);
        float clkdiv = clock_get_hz(clk_sys) / sm1_freq;
        auto sm1_prog_offs = pio_add_program(pio, &on_pulse_sm1_program);
        pio_on_pulse_sm1_init(pio, _pulse_length_sm, sm1_prog_offs, PIN_TX, clkdiv, _sampling_cfg.pulse_len_us / 10);
    }
    {
        auto sm3 = pio_claim_unused_sm(pio, true);
        _adc.init(pio, sm3);
    }
    {
        auto sm2 = pio_claim_unused_sm(pio, true);
        float clkdiv = 1.0f;
        auto sm2_prog_offs = pio_add_program(pio, &sample_counter_sm2_program);
        pio_sample_counter_sm2_init(pio, sm2, sm2_prog_offs, clkdiv, ADC_SAMPLE_DEPTH);
    }
}

void PISampler::checkSamplingConfigurationForPIO() {
    auto& pulse_len = _sampling_cfg.pulse_len_us;

    if (pulse_len < 10)
        pulse_len = 10;
    else if (pulse_len > 320)
        pulse_len = 320;

    //ensure divides by 10, round down if not
    pulse_len -= pulse_len % 10;

    auto& pulse_period = _sampling_cfg.pulse_period_us;
    if (pulse_period < 100)
        pulse_period = 100;
    if (pulse_period > 3200)
        pulse_period = 3200;

    //ensure divides by 100, round down if not

    pulse_period -= pulse_period % 100;

}

void PISampler::updatePIOSamplingConfiguration() {
    pio_sm_exec(_pio, _pulse_length_sm, pio_encode_set(pio_y, _sampling_cfg.pulse_len_us/10-1));
    pio_sm_exec(_pio, _pulse_period_sm, pio_encode_set(pio_y, _sampling_cfg.pulse_period_us/100-1));
}
